# -*- coding: utf-8 -*-
import json
import logging

log = logging.getLogger(__name__)


def is_network_appbase_ready(props):
    """
    Return True if the provided network/node properties indicate an appbase-ready node.

    Checks for the presence of the "HIVE_BLOCKCHAIN_VERSION" key in props.

    Parameters:
        props (Mapping): Mapping (e.g., dict) of network or node properties.

    Returns:
        bool: True if "HIVE_BLOCKCHAIN_VERSION" exists in props, otherwise False.
    """
    if "HIVE_BLOCKCHAIN_VERSION" in props:
        return True
    else:
        return False


def get_query(appbase, request_id, api_name, name, args):
    query = []
    if not appbase or api_name == "condenser_api":
        query = {
            "method": "call",
            "params": [api_name, name, list(args)],
            "jsonrpc": "2.0",
            "id": request_id,
        }
    else:
        args = json.loads(json.dumps(args))
        # print(args)
        if len(args) > 0 and isinstance(args, list) and isinstance(args[0], dict):
            query = {
                "method": api_name + "." + name,
                "params": args[0],
                "jsonrpc": "2.0",
                "id": request_id,
            }
        elif (
            len(args) > 0
            and isinstance(args, list)
            and isinstance(args[0], list)
            and len(args[0]) > 0
            and isinstance(args[0][0], dict)
        ):
            for a in args[0]:
                query.append(
                    {
                        "method": api_name + "." + name,
                        "params": a,
                        "jsonrpc": "2.0",
                        "id": request_id,
                    }
                )
                request_id += 1
        elif args:
            query = {
                "method": "call",
                "params": [api_name, name, list(args)],
                "jsonrpc": "2.0",
                "id": request_id,
            }
            request_id += 1
        elif api_name == "condenser_api":
            query = {
                "method": api_name + "." + name,
                "jsonrpc": "2.0",
                "params": [],
                "id": request_id,
            }
        else:
            query = {
                "method": api_name + "." + name,
                "jsonrpc": "2.0",
                "params": {},
                "id": request_id,
            }
    return query


def get_api_name(appbase, *args, **kwargs):
    if not appbase:
        # Sepcify the api to talk to
        if ("api" in kwargs) and len(kwargs["api"]) > 0:
            api_name = kwargs["api"].replace("_api", "") + "_api"
        else:
            api_name = None
    else:
        # Sepcify the api to talk to
        if ("api" in kwargs) and len(kwargs["api"]) > 0:
            if kwargs["api"] not in ["jsonrpc", "hive", "bridge"]:
                api_name = kwargs["api"].replace("_api", "") + "_api"
            else:
                api_name = kwargs["api"]
        else:
            api_name = "condenser_api"
    return api_name



# -*- coding: utf-8 -*-
# Inspired by https://raw.githubusercontent.com/xeroc/python-graphenelib/master/graphenestorage/exceptions.py
class WalletLocked(Exception):
    pass


class WrongMasterPasswordException(Exception):
    """The password provided could not properly unlock the wallet"""

    pass


class KeyAlreadyInStoreException(Exception):
    """The key of a key/value pair is already in the store"""

    pass



# -*- coding: utf-8 -*-
import json
import re
import warnings
from binascii import hexlify
from collections import OrderedDict

from nectargraphenebase.account import PublicKey
from nectargraphenebase.types import (
    Array,
    Bool,
    HexString,
    Int16,
    Map,
    Optional,
    PointInTime,
    String,
    Uint16,
    Uint32,
    Uint64,
)

from .objects import (
    Amount,
    CommentOptionExtensions,
    ExchangeRate,
    GrapheneObject,
    Memo,
    Operation,
    Permission,
    UpdateProposalExtensions,
    WitnessProps,
    isArgsThisClass,
)

default_prefix = "STM"


def check_for_class(self, args):
    if isArgsThisClass(self, args):
        self.data = args[0].data
        return True
    else:
        return False


class Transfer(GrapheneObject):
    def __init__(self, *args, **kwargs):
        # Allow for overwrite of prefix
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        if "memo" not in kwargs:
            kwargs["memo"] = ""
        if isinstance(kwargs["memo"], dict):
            kwargs["memo"]["prefix"] = prefix
            memo = Optional(Memo(**kwargs["memo"]))
        elif isinstance(kwargs["memo"], str):
            memo = String(kwargs["memo"])
        else:
            memo = Optional(Memo(kwargs["memo"]))

        super(Transfer, self).__init__(
            OrderedDict(
                [
                    ("from", String(kwargs["from"])),
                    ("to", String(kwargs["to"])),
                    ("amount", Amount(kwargs["amount"], prefix=prefix, json_str=json_str)),
                    ("memo", memo),
                ]
            )
        )


# Added recurring transfer support for HF25
class Recurring_transfer(GrapheneObject):
    def __init__(self, *args, **kwargs):
        # Allow for overwrite of prefix
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        if "memo" not in kwargs:
            kwargs["memo"] = ""
        if isinstance(kwargs["memo"], dict):
            kwargs["memo"]["prefix"] = prefix
            memo = Optional(Memo(**kwargs["memo"]))
        elif isinstance(kwargs["memo"], str):
            memo = String(kwargs["memo"])
        else:
            memo = Optional(Memo(kwargs["memo"]))

        super(Recurring_transfer, self).__init__(
            OrderedDict(
                [
                    ("from", String(kwargs["from"])),
                    ("to", String(kwargs["to"])),
                    ("amount", Amount(kwargs["amount"], prefix=prefix, json_str=json_str)),
                    ("memo", memo),
                    ("recurrence", Int16(kwargs["recurrence"])),
                    ("executions", Int16(kwargs["executions"])),
                ]
            )
        )


class Vote(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        super(Vote, self).__init__(
            OrderedDict(
                [
                    ("voter", String(kwargs["voter"])),
                    ("author", String(kwargs["author"])),
                    ("permlink", String(kwargs["permlink"])),
                    ("weight", Int16(kwargs["weight"])),
                ]
            )
        )


class Transfer_to_vesting(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        super(Transfer_to_vesting, self).__init__(
            OrderedDict(
                [
                    ("from", String(kwargs["from"])),
                    ("to", String(kwargs["to"])),
                    ("amount", Amount(kwargs["amount"], prefix=prefix, json_str=json_str)),
                ]
            )
        )


class Withdraw_vesting(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        super(Withdraw_vesting, self).__init__(
            OrderedDict(
                [
                    ("account", String(kwargs["account"])),
                    ("vesting_shares", Amount(kwargs["vesting_shares"], prefix=prefix)),
                ]
            )
        )


class Account_witness_vote(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        super(Account_witness_vote, self).__init__(
            OrderedDict(
                [
                    ("account", String(kwargs["account"])),
                    ("witness", String(kwargs["witness"])),
                    ("approve", Bool(bool(kwargs["approve"]))),
                ]
            )
        )


class Account_witness_proxy(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        super(Account_witness_proxy, self).__init__(
            OrderedDict(
                [
                    ("account", String(kwargs["account"])),
                    ("proxy", String(kwargs["proxy"])),
                ]
            )
        )


class Custom(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        super(Custom, self).__init__(
            OrderedDict(
                [
                    ("required_auths", Array([String(o) for o in kwargs["required_auths"]])),
                    ("id", Uint16(int(kwargs["id"]))),
                    ("data", String(kwargs["data"])),
                ]
            )
        )


class Custom_binary(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        super(Custom_binary, self).__init__(
            OrderedDict(
                [
                    ("id", Uint16(int(kwargs["id"]))),
                    ("data", String(kwargs["data"])),
                ]
            )
        )


class Op_wrapper(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        super(Op_wrapper, self).__init__(
            OrderedDict(
                [
                    ("op", Operation(kwargs["op"], prefix=prefix)),
                ]
            )
        )


class Account_create(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        if not len(kwargs["new_account_name"]) <= 16:
            raise AssertionError("Account name must be at most 16 chars long")

        meta = ""
        if "json_metadata" in kwargs and kwargs["json_metadata"]:
            if isinstance(kwargs["json_metadata"], dict):
                meta = json.dumps(kwargs["json_metadata"])
            else:
                meta = kwargs["json_metadata"]

        super(Account_create, self).__init__(
            OrderedDict(
                [
                    ("fee", Amount(kwargs["fee"], prefix=prefix, json_str=json_str)),
                    ("creator", String(kwargs["creator"])),
                    ("new_account_name", String(kwargs["new_account_name"])),
                    ("owner", Permission(kwargs["owner"], prefix=prefix)),
                    ("active", Permission(kwargs["active"], prefix=prefix)),
                    ("posting", Permission(kwargs["posting"], prefix=prefix)),
                    ("memo_key", PublicKey(kwargs["memo_key"], prefix=prefix)),
                    ("json_metadata", String(meta)),
                ]
            )
        )


class Account_create_with_delegation(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        if not len(kwargs["new_account_name"]) <= 16:
            raise AssertionError("Account name must be at most 16 chars long")

        meta = ""
        if "json_metadata" in kwargs and kwargs["json_metadata"]:
            if isinstance(kwargs["json_metadata"], dict):
                meta = json.dumps(kwargs["json_metadata"])
            else:
                meta = kwargs["json_metadata"]

        super(Account_create_with_delegation, self).__init__(
            OrderedDict(
                [
                    ("fee", Amount(kwargs["fee"], prefix=prefix, json_str=json_str)),
                    ("delegation", Amount(kwargs["delegation"], prefix=prefix, json_str=json_str)),
                    ("creator", String(kwargs["creator"])),
                    ("new_account_name", String(kwargs["new_account_name"])),
                    ("owner", Permission(kwargs["owner"], prefix=prefix)),
                    ("active", Permission(kwargs["active"], prefix=prefix)),
                    ("posting", Permission(kwargs["posting"], prefix=prefix)),
                    ("memo_key", PublicKey(kwargs["memo_key"], prefix=prefix)),
                    ("json_metadata", String(meta)),
                    ("extensions", Array([])),
                ]
            )
        )


class Account_update(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)

        if "owner" in kwargs:
            owner = Optional(Permission(kwargs["owner"], prefix=prefix))
        else:
            owner = Optional(None)

        if "active" in kwargs:
            active = Optional(Permission(kwargs["active"], prefix=prefix))
        else:
            active = Optional(None)

        if "posting" in kwargs:
            posting = Optional(Permission(kwargs["posting"], prefix=prefix))
        else:
            posting = Optional(None)

        meta = ""
        if "json_metadata" in kwargs and kwargs["json_metadata"]:
            if isinstance(kwargs["json_metadata"], dict):
                meta = json.dumps(kwargs["json_metadata"])
            else:
                meta = kwargs["json_metadata"]

        super(Account_update, self).__init__(
            OrderedDict(
                [
                    ("account", String(kwargs["account"])),
                    ("owner", owner),
                    ("active", active),
                    ("posting", posting),
                    ("memo_key", PublicKey(kwargs["memo_key"], prefix=prefix)),
                    ("json_metadata", String(meta)),
                ]
            )
        )


class Account_update2(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        extensions = Array([])

        if "owner" in kwargs:
            owner = Optional(Permission(kwargs["owner"], prefix=prefix))
        else:
            owner = Optional(None)

        if "active" in kwargs:
            active = Optional(Permission(kwargs["active"], prefix=prefix))
        else:
            active = Optional(None)

        if "posting" in kwargs:
            posting = Optional(Permission(kwargs["posting"], prefix=prefix))
        else:
            posting = Optional(None)

        if "memo_key" in kwargs:
            memo_key = Optional(PublicKey(kwargs["memo_key"], prefix=prefix))
        else:
            memo_key = Optional(None)

        meta = ""
        if "json_metadata" in kwargs and kwargs["json_metadata"]:
            if isinstance(kwargs["json_metadata"], dict):
                meta = json.dumps(kwargs["json_metadata"])
            else:
                meta = kwargs["json_metadata"]
        posting_meta = ""
        if "posting_json_metadata" in kwargs and kwargs["posting_json_metadata"]:
            if isinstance(kwargs["posting_json_metadata"], dict):
                posting_meta = json.dumps(kwargs["posting_json_metadata"])
            else:
                posting_meta = kwargs["posting_json_metadata"]

        super(Account_update2, self).__init__(
            OrderedDict(
                [
                    ("account", String(kwargs["account"])),
                    ("owner", owner),
                    ("active", active),
                    ("posting", posting),
                    ("memo_key", memo_key),
                    ("json_metadata", String(meta)),
                    ("posting_json_metadata", String(posting_meta)),
                    ("extensions", extensions),
                ]
            )
        )


class Create_proposal(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        extensions = Array([])

        super(Create_proposal, self).__init__(
            OrderedDict(
                [
                    ("creator", String(kwargs["creator"])),
                    ("receiver", String(kwargs["receiver"])),
                    ("start_date", PointInTime(kwargs["start_date"])),
                    ("end_date", PointInTime(kwargs["end_date"])),
                    ("daily_pay", Amount(kwargs["daily_pay"], prefix=prefix, json_str=json_str)),
                    ("subject", String(kwargs["subject"])),
                    ("permlink", String(kwargs["permlink"])),
                    ("extensions", extensions),
                ]
            )
        )


class Update_proposal_votes(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        extensions = Array([])
        proposal_ids = []
        for e in kwargs["proposal_ids"]:
            proposal_ids.append(Uint64(e))

        super(Update_proposal_votes, self).__init__(
            OrderedDict(
                [
                    ("voter", String(kwargs["voter"])),
                    ("proposal_ids", Array(proposal_ids)),
                    ("approve", Bool(kwargs["approve"])),
                    ("extensions", extensions),
                ]
            )
        )


class Remove_proposal(GrapheneObject):
    def __init__(self, *args, **kwargs):
        """
        Initialize a Remove_proposal operation.

        Creates the internal OrderedDict for a remove_proposal operation with:
        - proposal_owner: account name (String)
        - proposal_ids: list of Uint64-wrapped proposal IDs
        - extensions: empty Array

        If initialized with a single existing GrapheneObject instance, initialization returns early after copying that instance's data (handled by check_for_class).

        Required kwargs:
        - proposal_owner: str
        - proposal_ids: iterable of integers (each converted to Uint64)
        """
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        extensions = Array([])
        proposal_ids = []
        for e in kwargs["proposal_ids"]:
            proposal_ids.append(Uint64(e))

        super(Remove_proposal, self).__init__(
            OrderedDict(
                [
                    ("proposal_owner", String(kwargs["proposal_owner"])),
                    ("proposal_ids", Array(proposal_ids)),
                    ("extensions", extensions),
                ]
            )
        )


class Update_proposal(GrapheneObject):
    def __init__(self, *args, **kwargs):
        """
        Initialize an Update_proposal operation.

        Accepts either an existing Update_proposal instance (handled by check_for_class), a single positional dict, or keyword arguments. Required fields: `proposal_id`, `creator`, `daily_pay`, `subject`, and `permlink`. Optional `end_date` will be converted into an `update_proposal_end_date` extension. The `daily_pay` Amount uses the provided `prefix` kwarg if present, otherwise `default_prefix` is used.

        Accepted kwargs:
        - proposal_id: numeric id of the proposal (converted to Uint64)
        - creator: account name string (converted to String)
        - daily_pay: amount specifier (converted to Amount; honors `prefix`)
        - subject: short subject string (converted to String)
        - permlink: permlink string (converted to String)
        - end_date: optional datetime/string; if provided, added as an extension
        - prefix: optional asset/account prefix for Amount conversion (defaults to module `default_prefix`)

        No return value; constructs the internal OrderedDict representing the operation.
        """
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        extensions = Array([])
        if "end_date" in kwargs and kwargs["end_date"]:
            extension = {
                "type": "update_proposal_end_date",
                "value": {"end_date": kwargs["end_date"]},
            }
            extensions = Array([UpdateProposalExtensions(extension)])

        super(Update_proposal, self).__init__(
            OrderedDict(
                [
                    ("proposal_id", Uint64(kwargs["proposal_id"])),
                    ("creator", String(kwargs["creator"])),
                    (
                        "daily_pay",
                        Amount(kwargs["daily_pay"], prefix=kwargs.get("prefix", default_prefix)),
                    ),
                    ("subject", String(kwargs["subject"])),
                    ("permlink", String(kwargs["permlink"])),
                    ("extensions", extensions),
                ]
            )
        )


class Witness_set_properties(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.pop("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        extensions = Array([])
        props = {}
        for k in kwargs["props"]:
            if "key" == k[0]:
                block_signing_key = PublicKey(k[1], prefix=prefix)
                props["key"] = repr(block_signing_key)
            elif "new_signing_key" == k[0]:
                new_signing_key = PublicKey(k[1], prefix=prefix)
                props["new_signing_key"] = repr(new_signing_key)
        for k in kwargs["props"]:
            if k[0] in ["key", "new_signing_key"]:
                continue
            if isinstance(k[1], str):
                is_hex = re.fullmatch(r"[0-9a-fA-F]+", k[1] or "") is not None
            else:
                is_hex = False
            if isinstance(k[1], int) and k[0] in [
                "account_subsidy_budget",
                "account_subsidy_decay",
                "maximum_block_size",
            ]:
                props[k[0]] = (hexlify(Uint32(k[1]).__bytes__())).decode()
            elif isinstance(k[1], int) and k[0] in ["sbd_interest_rate", "hbd_interest_rate"]:
                props[k[0]] = (hexlify(Uint16(k[1]).__bytes__())).decode()
            elif not isinstance(k[1], str) and k[0] in ["account_creation_fee"]:
                props[k[0]] = (
                    hexlify(Amount(k[1], prefix=prefix, json_str=json_str).__bytes__())
                ).decode()
            elif not is_hex and isinstance(k[1], str) and k[0] in ["account_creation_fee"]:
                props[k[0]] = (
                    hexlify(Amount(k[1], prefix=prefix, json_str=json_str).__bytes__())
                ).decode()
            elif not isinstance(k[1], str) and k[0] in ["sbd_exchange_rate", "hbd_exchange_rate"]:
                if "prefix" not in k[1]:
                    k[1]["prefix"] = prefix
                props[k[0]] = (hexlify(ExchangeRate(k[1]).__bytes__())).decode()
            elif not is_hex and k[0] in ["url"]:
                props[k[0]] = (hexlify(String(k[1]).__bytes__())).decode()
            else:
                props[k[0]] = k[1]
        props_list = []
        for k in props:
            props_list.append(([String(k), HexString(props[k])]))
        props_list = sorted(
            props_list,
            key=lambda x: str(x[0]),
            reverse=False,
        )
        map_props = Map(props_list)

        super(Witness_set_properties, self).__init__(
            OrderedDict(
                [
                    ("owner", String(kwargs["owner"])),
                    ("props", map_props),
                    ("extensions", extensions),
                ]
            )
        )


class Witness_update(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.pop("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        if "block_signing_key" in kwargs and kwargs["block_signing_key"]:
            block_signing_key = PublicKey(kwargs["block_signing_key"], prefix=prefix)
        else:
            block_signing_key = PublicKey(
                prefix + "1111111111111111111111111111111114T1Anm", prefix=prefix
            )
        if "prefix" not in kwargs["props"]:
            kwargs["props"]["prefix"] = prefix

        super(Witness_update, self).__init__(
            OrderedDict(
                [
                    ("owner", String(kwargs["owner"])),
                    ("url", String(kwargs["url"])),
                    ("block_signing_key", block_signing_key),
                    ("props", WitnessProps(kwargs["props"])),
                    ("fee", Amount(kwargs["fee"], prefix=prefix, json_str=json_str)),
                ]
            )
        )


class Comment(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        meta = ""
        if "json_metadata" in kwargs and kwargs["json_metadata"]:
            if isinstance(kwargs["json_metadata"], dict) or isinstance(
                kwargs["json_metadata"], list
            ):
                meta = json.dumps(kwargs["json_metadata"])
            else:
                meta = kwargs["json_metadata"]

        super(Comment, self).__init__(
            OrderedDict(
                [
                    ("parent_author", String(kwargs["parent_author"])),
                    ("parent_permlink", String(kwargs["parent_permlink"])),
                    ("author", String(kwargs["author"])),
                    ("permlink", String(kwargs["permlink"])),
                    ("title", String(kwargs["title"])),
                    ("body", String(kwargs["body"])),
                    ("json_metadata", String(meta)),
                ]
            )
        )


class Custom_json(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        if "json" in kwargs and kwargs["json"]:
            if isinstance(kwargs["json"], dict) or isinstance(kwargs["json"], list):
                js = json.dumps(kwargs["json"], separators=(",", ":"))
            else:
                js = kwargs["json"]

        if len(kwargs["id"]) > 32:
            raise Exception("'id' too long")

        super(Custom_json, self).__init__(
            OrderedDict(
                [
                    ("required_auths", Array([String(o) for o in kwargs["required_auths"]])),
                    (
                        "required_posting_auths",
                        Array([String(o) for o in kwargs["required_posting_auths"]]),
                    ),
                    ("id", String(kwargs["id"])),
                    ("json", String(js)),
                ]
            )
        )


class Comment_options(GrapheneObject):
    def __init__(self, *args, **kwargs):
        """
        Initialize a Comment_options operation.

        This constructor builds the serialized fields for a comment options operation from provided keyword arguments or a single dict positional argument. It converts and validates inputs into the expected Graphene types and handles backward compatibility and extensions.

        Expected kwargs:
        - author (str): post author.
        - permlink (str): post permlink.
        - max_accepted_payout (str|Amount): payout limit; converted to Amount using optional `prefix` and `json_str`.
        - percent_hbd (int|str): required percent value (primary source) stored as Uint16.
        - percent_steem_dollars (int|str, optional, deprecated): fallback for `percent_hbd`; using it emits a DeprecationWarning.
        - allow_votes (bool): whether voting is allowed.
        - allow_curation_rewards (bool): whether curation rewards are allowed.
        - beneficiaries (list, optional): if provided, placed into extensions as a beneficiaries extension.
        - extensions (iterable, optional): explicit extensions; each entry is wrapped with CommentOptionExtensions.
        - prefix (str, optional): asset/account prefix used when constructing Amount (defaults to module default_prefix).
        - json_str (bool, optional): if true, construct Amount with json string mode.

        Behavior and side effects:
        - If initialized from an existing GrapheneObject (via check_for_class), initialization returns early after copying.
        - If `beneficiaries` is present and non-empty, it is converted into an extensions entry.
        - If neither `percent_hbd` nor `percent_steem_dollars` is provided, raises ValueError.
        - If `percent_steem_dollars` is used, emits a DeprecationWarning.
        """
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        # handle beneficiaries
        if "beneficiaries" in kwargs and kwargs["beneficiaries"]:
            kwargs["extensions"] = [[0, {"beneficiaries": kwargs["beneficiaries"]}]]
        extensions = Array([])
        if "extensions" in kwargs and kwargs["extensions"]:
            extensions = Array([CommentOptionExtensions(o) for o in kwargs["extensions"]])
        percent_value = kwargs.get("percent_hbd")
        if percent_value is None:
            # Backward compatibility: fall back to percent_steem_dollars
            percent_value = kwargs.get("percent_steem_dollars")
            if percent_value is not None:
                warnings.warn(
                    "Parameter 'percent_steem_dollars' is deprecated. Use 'percent_hbd' instead.",
                    DeprecationWarning,
                    stacklevel=2,
                )
            else:
                raise ValueError("Comment_options requires 'percent_hbd'")
        super(Comment_options, self).__init__(
            OrderedDict(
                [
                    ("author", String(kwargs["author"])),
                    ("permlink", String(kwargs["permlink"])),
                    (
                        "max_accepted_payout",
                        Amount(kwargs["max_accepted_payout"], prefix=prefix, json_str=json_str),
                    ),
                    ("percent_hbd", Uint16(int(percent_value))),
                    ("allow_votes", Bool(bool(kwargs["allow_votes"]))),
                    ("allow_curation_rewards", Bool(bool(kwargs["allow_curation_rewards"]))),
                    ("extensions", extensions),
                ]
            )
        )


class Delete_comment(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        super(Delete_comment, self).__init__(
            OrderedDict(
                [
                    ("author", String(kwargs["author"])),
                    ("permlink", String(kwargs["permlink"])),
                ]
            )
        )


class Feed_publish(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        if "prefix" not in kwargs["exchange_rate"]:
            kwargs["exchange_rate"]["prefix"] = prefix
        super(Feed_publish, self).__init__(
            OrderedDict(
                [
                    ("publisher", String(kwargs["publisher"])),
                    ("exchange_rate", ExchangeRate(kwargs["exchange_rate"])),
                ]
            )
        )


class Convert(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        super(Convert, self).__init__(
            OrderedDict(
                [
                    ("owner", String(kwargs["owner"])),
                    ("requestid", Uint32(kwargs["requestid"])),
                    ("amount", Amount(kwargs["amount"], prefix=prefix, json_str=json_str)),
                ]
            )
        )


# Operation added for HF25 for the new HBD/Hive conversion operation
class Collateralized_convert(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        super(Collateralized_convert, self).__init__(
            OrderedDict(
                [
                    ("owner", String(kwargs["owner"])),
                    ("requestid", Uint32(kwargs["requestid"])),
                    ("amount", Amount(kwargs["amount"], prefix=prefix, json_str=json_str)),
                ]
            )
        )


class Set_withdraw_vesting_route(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        super(Set_withdraw_vesting_route, self).__init__(
            OrderedDict(
                [
                    ("from_account", String(kwargs["from_account"])),
                    ("to_account", String(kwargs["to_account"])),
                    ("percent", Uint16((kwargs["percent"]))),
                    ("auto_vest", Bool(kwargs["auto_vest"])),
                ]
            )
        )


class Limit_order_cancel(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        super(Limit_order_cancel, self).__init__(
            OrderedDict(
                [
                    ("owner", String(kwargs["owner"])),
                    ("orderid", Uint32(kwargs["orderid"])),
                ]
            )
        )


class Claim_account(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        super(Claim_account, self).__init__(
            OrderedDict(
                [
                    ("creator", String(kwargs["creator"])),
                    ("fee", Amount(kwargs["fee"], prefix=prefix, json_str=json_str)),
                    ("extensions", Array([])),
                ]
            )
        )


class Create_claimed_account(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)

        if not len(kwargs["new_account_name"]) <= 16:
            raise AssertionError("Account name must be at most 16 chars long")

        meta = ""
        if "json_metadata" in kwargs and kwargs["json_metadata"]:
            if isinstance(kwargs["json_metadata"], dict):
                meta = json.dumps(kwargs["json_metadata"])
            else:
                meta = kwargs["json_metadata"]

        super(Create_claimed_account, self).__init__(
            OrderedDict(
                [
                    ("creator", String(kwargs["creator"])),
                    ("new_account_name", String(kwargs["new_account_name"])),
                    ("owner", Permission(kwargs["owner"], prefix=prefix)),
                    ("active", Permission(kwargs["active"], prefix=prefix)),
                    ("posting", Permission(kwargs["posting"], prefix=prefix)),
                    ("memo_key", PublicKey(kwargs["memo_key"], prefix=prefix)),
                    ("json_metadata", String(meta)),
                    ("extensions", Array([])),
                ]
            )
        )


class Delegate_vesting_shares(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        super(Delegate_vesting_shares, self).__init__(
            OrderedDict(
                [
                    ("delegator", String(kwargs["delegator"])),
                    ("delegatee", String(kwargs["delegatee"])),
                    ("vesting_shares", Amount(kwargs["vesting_shares"], prefix=prefix)),
                ]
            )
        )


class Limit_order_create(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        super(Limit_order_create, self).__init__(
            OrderedDict(
                [
                    ("owner", String(kwargs["owner"])),
                    ("orderid", Uint32(kwargs["orderid"])),
                    (
                        "amount_to_sell",
                        Amount(kwargs["amount_to_sell"], prefix=prefix, json_str=json_str),
                    ),
                    (
                        "min_to_receive",
                        Amount(kwargs["min_to_receive"], prefix=prefix, json_str=json_str),
                    ),
                    ("fill_or_kill", Bool(kwargs["fill_or_kill"])),
                    ("expiration", PointInTime(kwargs["expiration"])),
                ]
            )
        )


class Limit_order_create2(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        if "prefix" not in kwargs["exchange_rate"]:
            kwargs["exchange_rate"]["prefix"] = prefix
        super(Limit_order_create2, self).__init__(
            OrderedDict(
                [
                    ("owner", String(kwargs["owner"])),
                    ("orderid", Uint32(kwargs["orderid"])),
                    (
                        "amount_to_sell",
                        Amount(kwargs["amount_to_sell"], prefix=prefix, json_str=json_str),
                    ),
                    ("fill_or_kill", Bool(kwargs["fill_or_kill"])),
                    ("exchange_rate", ExchangeRate(kwargs["exchange_rate"])),
                    ("expiration", PointInTime(kwargs["expiration"])),
                ]
            )
        )


class Change_recovery_account(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        super(Change_recovery_account, self).__init__(
            OrderedDict(
                [
                    ("account_to_recover", String(kwargs["account_to_recover"])),
                    ("new_recovery_account", String(kwargs["new_recovery_account"])),
                    ("extensions", Array([])),
                ]
            )
        )


class Transfer_from_savings(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        if "memo" not in kwargs:
            kwargs["memo"] = ""
        super(Transfer_from_savings, self).__init__(
            OrderedDict(
                [
                    ("from", String(kwargs["from"])),
                    ("request_id", Uint32(kwargs["request_id"])),
                    ("to", String(kwargs["to"])),
                    ("amount", Amount(kwargs["amount"], prefix=prefix, json_str=json_str)),
                    ("memo", String(kwargs["memo"])),
                ]
            )
        )


class Cancel_transfer_from_savings(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        super(Cancel_transfer_from_savings, self).__init__(
            OrderedDict(
                [
                    ("from", String(kwargs["from"])),
                    ("request_id", Uint32(kwargs["request_id"])),
                ]
            )
        )


class Claim_reward_balance(GrapheneObject):
    def __init__(self, *args, **kwargs):
        """
        Initialize a Claim_reward_balance operation.

        Constructs the serialized fields for claiming reward balances. Requires
        account, reward_hive, reward_hbd, and reward_vests in the canonical order.
        All reward fields are required asset strings - use "0.000 HIVE" or "0.000 HBD"
        when nothing to claim for that asset.

        Behavior:
        - Always serializes ("account", "reward_hive", "reward_hbd", "reward_vests")
        - Converts provided values to Amount objects, respecting prefix/json_str behavior
        - Uses zero-asset strings ("0.000 HIVE"/"0.000 HBD") for any missing reward fields

        Recognized kwargs:
        - account (str): account name claiming rewards.
        - reward_hive (str|Amount): HIVE amount to claim (required, use "0.000 HIVE" if none).
        - reward_hbd (str|Amount): HBD amount to claim (required, use "0.000 HBD" if none).
        - reward_vests (str|Amount): VESTS amount to claim.
        - prefix (str): asset prefix to use (defaults to module default_prefix).
        - json_str (bool): if True, pass amounts as JSON-string form to Amount.

        Also supports initialization from an existing instance via the module's check_for_class helper.
        """
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)

        # Ensure all required fields are present, using zero amounts for missing rewards
        account = kwargs["account"]
        reward_hive = kwargs.get("reward_hive", "0.000 HIVE")
        reward_hbd = kwargs.get("reward_hbd", "0.000 HBD")
        reward_vests = kwargs["reward_vests"]

        super(Claim_reward_balance, self).__init__(
            OrderedDict(
                [
                    ("account", String(account)),
                    ("reward_hive", Amount(reward_hive, prefix=prefix, json_str=json_str)),
                    ("reward_hbd", Amount(reward_hbd, prefix=prefix, json_str=json_str)),
                    ("reward_vests", Amount(reward_vests, prefix=prefix, json_str=json_str)),
                ]
            )
        )


class Transfer_to_savings(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        if "memo" not in kwargs:
            kwargs["memo"] = ""
        super(Transfer_to_savings, self).__init__(
            OrderedDict(
                [
                    ("from", String(kwargs["from"])),
                    ("to", String(kwargs["to"])),
                    ("amount", Amount(kwargs["amount"], prefix=prefix, json_str=json_str)),
                    ("memo", String(kwargs["memo"])),
                ]
            )
        )


class Request_account_recovery(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        new_owner = Permission(kwargs["new_owner_authority"], prefix=prefix)
        super(Request_account_recovery, self).__init__(
            OrderedDict(
                [
                    ("recovery_account", String(kwargs["recovery_account"])),
                    ("account_to_recover", String(kwargs["account_to_recover"])),
                    ("new_owner_authority", new_owner),
                    ("extensions", Array([])),
                ]
            )
        )


class Recover_account(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        new_owner = Permission(kwargs["new_owner_authority"], prefix=prefix)
        recent_owner = Permission(kwargs["recent_owner_authority"], prefix=prefix)
        super(Recover_account, self).__init__(
            OrderedDict(
                [
                    ("account_to_recover", String(kwargs["account_to_recover"])),
                    ("new_owner_authority", new_owner),
                    ("recent_owner_authority", recent_owner),
                    ("extensions", Array([])),
                ]
            )
        )


class Escrow_transfer(GrapheneObject):
    def __init__(self, *args, **kwargs):
        """
        Initialize an Escrow_transfer operation object.

        If constructed from an existing GrapheneObject instance (detected via check_for_class), the initializer returns early after copying data.

        Accepts either a single dict positional argument or keyword arguments. Expected fields:
        - from, to, agent (str): account names involved.
        - escrow_id (int): escrow identifier.
        - hbd_amount, hive_amount, fee: amounts; when both `hbd_amount` and `hive_amount` are provided, amounts are wrapped with the `json_str` option; otherwise amounts are wrapped without `json_str`.
        - ratification_deadline, escrow_expiration: datetime-like values for deadlines.
        - json_meta: optional metadata — if a dict or list it will be JSON-serialized; otherwise used as-is.
        Optional kwargs:
        - prefix (str): asset prefix (default "STM").
        - json_str (bool): whether to force JSON string representation for Amount fields when the branch requires it.

        No return value; constructs and initializes the underlying ordered field mapping for the operation.
        """
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        meta = ""
        if "json_meta" in kwargs and kwargs["json_meta"]:
            if isinstance(kwargs["json_meta"], dict) or isinstance(kwargs["json_meta"], list):
                meta = json.dumps(kwargs["json_meta"])
            else:
                meta = kwargs["json_meta"]
        if "hbd_amount" in kwargs and "hive_amount" in kwargs:
            super(Escrow_transfer, self).__init__(
                OrderedDict(
                    [
                        ("from", String(kwargs["from"])),
                        ("to", String(kwargs["to"])),
                        ("agent", String(kwargs["agent"])),
                        ("escrow_id", Uint32(kwargs["escrow_id"])),
                        (
                            "hbd_amount",
                            Amount(kwargs["hbd_amount"], prefix=prefix, json_str=json_str),
                        ),
                        (
                            "hive_amount",
                            Amount(kwargs["hive_amount"], prefix=prefix, json_str=json_str),
                        ),
                        ("fee", Amount(kwargs["fee"], prefix=prefix, json_str=json_str)),
                        ("ratification_deadline", PointInTime(kwargs["ratification_deadline"])),
                        ("escrow_expiration", PointInTime(kwargs["escrow_expiration"])),
                        ("json_meta", String(meta)),
                    ]
                )
            )
        else:
            super(Escrow_transfer, self).__init__(
                OrderedDict(
                    [
                        ("from", String(kwargs["from"])),
                        ("to", String(kwargs["to"])),
                        ("agent", String(kwargs["agent"])),
                        ("escrow_id", Uint32(kwargs["escrow_id"])),
                        ("hbd_amount", Amount(kwargs["hbd_amount"], prefix=prefix)),
                        ("hive_amount", Amount(kwargs["hive_amount"], prefix=prefix)),
                        ("fee", Amount(kwargs["fee"], prefix=prefix)),
                        ("ratification_deadline", PointInTime(kwargs["ratification_deadline"])),
                        ("escrow_expiration", PointInTime(kwargs["escrow_expiration"])),
                        ("json_meta", String(meta)),
                    ]
                )
            )


class Escrow_dispute(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        super(Escrow_dispute, self).__init__(
            OrderedDict(
                [
                    ("from", String(kwargs["from"])),
                    ("to", String(kwargs["to"])),
                    ("who", String(kwargs["who"])),
                    ("escrow_id", Uint32(kwargs["escrow_id"])),
                ]
            )
        )


class Escrow_release(GrapheneObject):
    def __init__(self, *args, **kwargs):
        """
        Initialize an Escrow_release operation.

        Constructs the operation fields required to release escrowed funds: from, to, who, escrow_id, hbd_amount, and hive_amount. Accepts either a single dict positional argument or keyword arguments. If initialized from an existing GrapheneObject instance (detected by check_for_class), initialization returns early after cloning.

        Key kwargs:
        - from, to, who (str): account names involved in the escrow release.
        - escrow_id (int): escrow identifier.
        - hbd_amount, hive_amount (str|Amount): amounts to release; wrapped as Amount objects using the provided prefix.
        - prefix (str, optional): asset/account prefix passed to Amount (defaults to default_prefix).
        - json_str (bool, optional): when True and both amount keys are present, amounts are wrapped with json_str enabled.

        Raises:
        - KeyError if any required field is missing.
        """
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        prefix = kwargs.get("prefix", default_prefix)
        json_str = kwargs.get("json_str", False)
        if "hive_amount" in kwargs and "hbd_amount" in kwargs:
            super(Escrow_release, self).__init__(
                OrderedDict(
                    [
                        ("from", String(kwargs["from"])),
                        ("to", String(kwargs["to"])),
                        ("who", String(kwargs["who"])),
                        ("escrow_id", Uint32(kwargs["escrow_id"])),
                        (
                            "hbd_amount",
                            Amount(kwargs["hbd_amount"], prefix=prefix, json_str=json_str),
                        ),
                        (
                            "hive_amount",
                            Amount(kwargs["hive_amount"], prefix=prefix, json_str=json_str),
                        ),
                    ]
                )
            )
        else:
            super(Escrow_release, self).__init__(
                OrderedDict(
                    [
                        ("from", String(kwargs["from"])),
                        ("to", String(kwargs["to"])),
                        ("who", String(kwargs["who"])),
                        ("escrow_id", Uint32(kwargs["escrow_id"])),
                        ("hbd_amount", Amount(kwargs["hbd_amount"], prefix=prefix)),
                        ("hive_amount", Amount(kwargs["hive_amount"], prefix=prefix)),
                    ]
                )
            )


class Escrow_approve(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        super(Escrow_approve, self).__init__(
            OrderedDict(
                [
                    ("from", String(kwargs["from"])),
                    ("to", String(kwargs["to"])),
                    ("agent", String(kwargs["agent"])),
                    ("who", String(kwargs["who"])),
                    ("escrow_id", Uint32(kwargs["escrow_id"])),
                    ("approve", Bool(kwargs["approve"])),
                ]
            )
        )


class Decline_voting_rights(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if check_for_class(self, args):
            return
        if len(args) == 1 and len(kwargs) == 0:
            kwargs = args[0]
        super(Decline_voting_rights, self).__init__(
            OrderedDict(
                [
                    ("account", String(kwargs["account"])),
                    ("decline", Bool(kwargs["decline"])),
                ]
            )
        )



# -*- coding: utf-8 -*-
#!/usr/bin/python
import unittest
from datetime import datetime, timedelta, timezone

from nectar import Hive, exceptions
from nectar.account import Account
from nectar.comment import Comment
from nectar.instance import set_shared_blockchain_instance
from nectar.utils import (
    construct_authorperm,
    construct_authorpermvoter,
    resolve_authorpermvoter,
)
from nectar.vote import AccountVotes, ActiveVotes, Vote

from .nodes import get_hive_nodes

wif = "5KQwrPbwdL6PhXujxW37FSSQZ1JiwsST4cqQzDeyXtP79zkvFD3"


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.bts = Hive(
            node=get_hive_nodes(), nobroadcast=True, keys={"active": wif}, num_retries=10
        )
        # from getpass import getpass
        # self.bts.wallet.unlock(getpass())
        set_shared_blockchain_instance(cls.bts)
        cls.bts.set_default_account("test")

        acc = Account("fullnodeupdate", blockchain_instance=cls.bts)
        n_votes = 0
        index = 0
        entries = acc.get_blog(limit=20)[::-1]
        while n_votes == 0:
            comment = Comment(entries[index], blockchain_instance=cls.bts)
            votes = comment.get_votes()
            n_votes = len(votes)
            index += 1

        last_vote = votes[0]

        cls.authorpermvoter = construct_authorpermvoter(
            last_vote["author"], last_vote["permlink"], last_vote["voter"]
        )
        [author, permlink, voter] = resolve_authorpermvoter(cls.authorpermvoter)
        cls.author = author
        cls.permlink = permlink
        cls.voter = voter
        cls.authorperm = construct_authorperm(author, permlink)

    def test_vote(self):
        bts = self.bts
        self.assertTrue(len(bts.get_reward_funds(use_stored_data=False)) > 0)
        vote = Vote(self.authorpermvoter, blockchain_instance=bts)
        self.assertEqual(self.voter, vote["voter"])
        self.assertEqual(self.author, vote["author"])
        self.assertEqual(self.permlink, vote["permlink"])

        vote = Vote(self.voter, authorperm=self.authorperm, blockchain_instance=bts)
        self.assertEqual(self.voter, vote["voter"])
        self.assertEqual(self.author, vote["author"])
        self.assertEqual(self.permlink, vote["permlink"])
        vote_json = vote.json()
        self.assertEqual(self.voter, vote_json["voter"])
        self.assertEqual(self.voter, vote.voter)
        self.assertTrue(vote.weight >= 0)
        if vote.percent >= 0:
            self.assertTrue(vote.hbd >= 0)
            self.assertTrue(vote.rshares >= 0)
        else:
            self.assertTrue(vote.hbd < 0)
            self.assertTrue(vote.rshares < 0)

        self.assertTrue(vote.reputation is not None)
        self.assertTrue(vote.rep is not None)
        self.assertTrue(vote.time is not None)
        vote.refresh()
        self.assertEqual(self.voter, vote["voter"])
        self.assertEqual(self.author, vote["author"])
        self.assertEqual(self.permlink, vote["permlink"])
        vote_json = vote.json()
        self.assertEqual(self.voter, vote_json["voter"])
        self.assertEqual(self.voter, vote.voter)
        self.assertTrue(vote.weight >= 0)
        if vote.percent >= 0:
            self.assertTrue(vote.hbd >= 0)
            self.assertTrue(vote.rshares >= 0)
        else:
            self.assertTrue(vote.hbd < 0)
            self.assertTrue(vote.rshares < 0)
        self.assertTrue(vote.reputation is not None)
        self.assertTrue(vote.rep is not None)
        self.assertTrue(vote.time is not None)

    def test_keyerror(self):
        bts = self.bts
        with self.assertRaises(exceptions.VoteDoesNotExistsException):
            Vote(
                construct_authorpermvoter(self.author, self.permlink, "asdfsldfjlasd"),
                blockchain_instance=bts,
            )

        with self.assertRaises(exceptions.VoteDoesNotExistsException):
            Vote(
                construct_authorpermvoter(self.author, "sdlfjd", "asdfsldfjlasd"),
                blockchain_instance=bts,
            )

        with self.assertRaises(exceptions.VoteDoesNotExistsException):
            Vote(
                construct_authorpermvoter("sdalfj", "dsfa", "asdfsldfjlasd"),
                blockchain_instance=bts,
            )

    def test_activevotes(self):
        bts = self.bts
        votes = ActiveVotes(self.authorperm, blockchain_instance=bts)
        votes.printAsTable()
        vote_list = votes.get_list()
        self.assertTrue(isinstance(vote_list, list))

    @unittest.skip
    def test_accountvotes(self):
        bts = self.bts
        limit_time = datetime.now(timezone.utc) - timedelta(days=7)
        votes = AccountVotes(self.voter, start=limit_time, blockchain_instance=bts)
        self.assertTrue(len(votes) > 0)
        self.assertTrue(isinstance(votes[0], Vote))



# This Python file uses the following encoding: utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import unittest
from builtins import bytes, chr, range
from itertools import cycle

from nectarbase.memo import (
    _pad,
    _unpad,
    decode_memo,
    decode_memo_bts,
    encode_memo,
    encode_memo_bts,
    get_shared_secret,
)
from nectargraphenebase.account import PasswordKey, PrivateKey, PublicKey

test_cases = [
    {
        "from": "GPH7FPzbN7hnRk24T3Nh9MYM1xBaF5xyRYu8WtyTrtLoUG8cUtszM",
        "message_bts": "688fe6c97f78ad2d3c5a82d9aa61bc23",
        "message": "#FYu8pMPJxTv7q2geNLSQC8dm47uqdNtFLCoDY5yZWjAz2R4wNyHEwQ48hPWm9SuAZ6fCFmjQrFCBVQFSP7EkobrWWRGaeqH6msKkPjRsMd6UUaNva1nmtLc55RAzqPLht",
        "nonce": "16332877645293003478",
        "plain": "I am this!",
        "to": "GPH6HAMuJRkjGJkj6cZWBbTU13gkUhBep383prqRdExXsZsYTrWT5",
        "wif": "5Jpkeq1jiNE8Pe24GxFWTsyWbcP59Qq4cD7qg3Wgd6JFJqJkoG8",
    },
    {  # no with integer nonce
        "from": "GPH7FPzbN7hnRk24T3Nh9MYM1xBaF5xyRYu8WtyTrtLoUG8cUtszM",
        "message_bts": "688fe6c97f78ad2d3c5a82d9aa61bc23",
        "message": "#FYu8pMPJxTv7q2geNLSQC8dm47uqdNtFLCoDY5yZWjAz2R4wNyHEwQ48hPWm9SuAZ6fCFmjQrFCBVQFSP7EkobrWWRGaeqH6msKkPjRsMd6UUaNva1nmtLc55RAzqPLht",
        "nonce": 16332877645293003478,
        "plain": "I am this!",
        "to": "GPH6HAMuJRkjGJkj6cZWBbTU13gkUhBep383prqRdExXsZsYTrWT5",
        "wif": "5Jpkeq1jiNE8Pe24GxFWTsyWbcP59Qq4cD7qg3Wgd6JFJqJkoG8",
    },
    {
        "from": "GPH7FPzbN7hnRk24T3Nh9MYM1xBaF5xyRYu8WtyTrtLoUG8cUtszM",
        "message_bts": "db7ab7dfefee3ffa2394ec438601ceff",
        "message": "#FYu8pMPJxTv7q2geNLSQC8dm47uqdNtFLCoDY5yZWjAz2R4wNyHEwQ48hPWm9SuAZ6fCFmjQrFCBVQFSP7EkobrWWRGaeqH6msKkPjRsMd6pNxowQQGhkWuR9z5W1aLau",
        "nonce": "16332877645293003478",
        "plain": "Hello World",
        "to": "GPH6HAMuJRkjGJkj6cZWBbTU13gkUhBep383prqRdExXsZsYTrWT5",
        "wif": "5Jpkeq1jiNE8Pe24GxFWTsyWbcP59Qq4cD7qg3Wgd6JFJqJkoG8",
    },
    {
        "from": "GPH7FPzbN7hnRk24T3Nh9MYM1xBaF5xyRYu8WtyTrtLoUG8cUtszM",
        "message_bts": "01b6616cbd10bdd0743c82c2bd580651f3e852360a739e7d11c45f483871dc45",
        "message": "#FYu8pMPJxTv7q2geNLSQC8dm47uqdNtFLCoDY5yZWjAz2R4wNyHEwQ48hPWm9SuAZ6fCFmjQrFCBVQFSP7EkobrWWRGaeqH6msKkPjRsMd6iKUwipf3H34zh3CAZVHNDy",
        "nonce": "16332877645293003478",
        "plain": "Daniel Larimer",
        "to": "GPH6HAMuJRkjGJkj6cZWBbTU13gkUhBep383prqRdExXsZsYTrWT5",
        "wif": "5Jpkeq1jiNE8Pe24GxFWTsyWbcP59Qq4cD7qg3Wgd6JFJqJkoG8",
    },
    {
        "from": "GPH7FPzbN7hnRk24T3Nh9MYM1xBaF5xyRYu8WtyTrtLoUG8cUtszM",
        "message_bts": "24702af49bc82e06eb74a4acd91b18c389b13a6c9850a0fd3f728f486fe6daf4",
        "message": "#8vxJp5YDC1Mv7J8sShbhdyrDNyo2JFuUxMmkYvg3tREpXDxoAvZSxzJ8Yqhx6qCyKfpHVczST9ySdXQANy2XBdFpztTu29pUibJBUzoKWgKYQyn7ixqUKhkexUA9Vt7W4crzbvnHhoB9Xogj9xxyhiN",
        "nonce": "16332877645293003478",
        "plain": "Thanks you, sir!",
        "to": "GPH6HAMuJRkjGJkj6cZWBbTU13gkUhBep383prqRdExXsZsYTrWT5",
        "wif": "5Jpkeq1jiNE8Pe24GxFWTsyWbcP59Qq4cD7qg3Wgd6JFJqJkoG8",
    },
    {
        "from": "GPH7FPzbN7hnRk24T3Nh9MYM1xBaF5xyRYu8WtyTrtLoUG8cUtszM",
        "message_bts": "1566da5b57e8e0fd9f530a352812a4197b8113df6495efdb246909c6ee1ffea6",
        "message": "#8vxJp5YDC1Mv7J8sShbhdyrDNyo2JFuUxMmkYvg3tREpXDxoAvZSxzJ8Yqhx6qCyKfpHVczST9ySdXQANy2XBdFpztTu29pUibJBUzoKWgKfW6Ew2qCiLrZFodoFmpQ77fg7HRGRRLN42jkJs3HxJf1",
        "nonce": "16332877645293003478",
        "plain": "äöüß€@$²³",
        "to": "GPH6HAMuJRkjGJkj6cZWBbTU13gkUhBep383prqRdExXsZsYTrWT5",
        "wif": "5Jpkeq1jiNE8Pe24GxFWTsyWbcP59Qq4cD7qg3Wgd6JFJqJkoG8",
    },
    {
        "from": "GPH6APYcWtrWXBhcrjPEhPz41bc98NxjnvufVVnRH1M8sjwtvFacz",
        "message_bts": "40b7ed2efd5e23b97e3f3aec6319fda722194e08b4cee45b84566e2741916797",
        "message": "#D2BAH3MLo3eMbJh9nR5jy53KXf22b55fQpNLXoGD4bqkE3EkiZirwL8GWsaFJ6g1RDzgRXiYXuNFwCyDddHzuL1Sxam5KCEMYZY4E5MmvMnv46ptN1Bur7Yuo7X6tfRtU",
        "nonce": "10864609094208714729",
        "plain": "1234567890\x02\x02",  # final bytes LOOK LIKE padding
        "to": "GPH7Ge953jTDzHKxFAzy19uhJtXxw8CbBM938hkSKWE3yXfRjLV57",
        "wif": "5KR8jzysz2kbYy3TkL3x6NRxfNXwQUWyeVAF5ZagxdqKMawGgXG",
    },
    {
        "from": "GPH6APYcWtrWXBhcrjPEhPz41bc98NxjnvufVVnRH1M8sjwtvFacz",
        "message_bts": "f43800298f9974c7b334bb1bf6224f236309520e99697f3980775231bfb4ef21",
        "message": "#D2BAH3MLo3eMbJh9nR5jy53KXf22b55fQpNLXoGD4bqkE3EkiZirwL8GWsaFJ6g1RDzgRXiYXuNFwCyDddHzuL1SxTXsyHkXiqBXGwC9v8guy8xFQQ7w5dLFXVHHgmZSV",
        "nonce": "8555724032490455626",
        "plain": "abcdefghijÛ",  # padding limit and last character is unicode
        "to": "GPH7Ge953jTDzHKxFAzy19uhJtXxw8CbBM938hkSKWE3yXfRjLV57",
        "wif": "5KR8jzysz2kbYy3TkL3x6NRxfNXwQUWyeVAF5ZagxdqKMawGgXG",
    },
]

test_shared_secrets = [
    [
        "5JYWCqDpeVrefVaFxJfDc3mzQ67dtsfhU7zcB7AMJYuTH57VsoE",
        "GPH56EzLTXkis55hBsompVXmSdnayG3afDNFmsCLohPh6rSNzkzhs",
        "fbb2fef5a3a115887df84c694e8ac5c9bf998c89d0c22438c18fd018f2529460",
    ],
    [
        "5JKhu9ZKydGFz7yGURocDVEepSY9fk2VRGAA8Xnb9wwFWa8yTWy",
        "GPH818iy2auxecLxhWTtW219w2VAfBYHxeHaeRASoTFLnsZo1DJ63",
        "4a52093355abeb31cef02ee1cbdf0661d982d52ad8fe39c68957e3ae03f3bda9",
    ],
    [
        "5KKmTkFCNnedj6hbyRYJwcaMnc4TkuwrPsJDqR2Bj9ShHkfdgQ3",
        "GPH78SdnBpqhEHxxzwZeKoFEXV6PviymWzBF7ev29pZcTCF8ynJAo",
        "500e67a07f53d49b88db635c64e4b0a2414168c7054118d40001e86f1abce131",
    ],
    [
        "5KLBuZtagfmGqhDTEPSM84TXKxKfzNyGaxRKCgdcocEU7Nusw49",
        "GPH5nYv9AusGXgHyMBbSBV4HyEAmhzXqLNRPvUpKmNpFo5soho95o",
        "febfa7ad6c48bb0ab976c6416da24017b93a58e4e699dba76fc590b4b1ac0d26",
    ],
    [
        "5JrVxMdeBZJvWqV4SmyFq9psQ4Dg8cFXtSWDiL7V5gUJC133xC2",
        "GPH7HxVNixmh33R44Kr2uJERbhvzkaLen8su4juqyFe2FW2U2cCXA",
        "82ef43913f83dd3ff0b4f06bcd8801a06c9f046b44b054e0a9ad042c28e5bdba",
    ],
    [
        "5JfEonXJ4H2kSP4V9NzC3uTRtTpLx4wVgDvf5AWN1KKTV6CZ4x7",
        "GPH6ZtaoP6skA433YGNNJcPGnsgx15psKRBwAy83tw7XWsDy8hso3",
        "b1ad058e9cc48e305fb46f07736409a55692c67d3507aad6a051b35459ec2f93",
    ],
    [
        "5J5UDLdk9XjvcbzNY5AQoUB2pttsvN7FtQFyyFZXUUsHFAp9iQd",
        "GPH51wPrJXWLcX6iNPAoZ9sGk4fHXk6krQgTX1jfuyxtKuhoEan83",
        "82fcc73de1331913945f6ce6d0207864bbc7cffee10ed3533ab32629cb759323",
    ],
    [
        "5JNZpagkR8wWsW3n4hHqFUQVkAu3HhJ9kU1criuruFpAwoaesHs",
        "GPH5SCy1teB91pNYetxEwV8vRyMApsy8aG61wsi8z2B4Zb6kfnqUf",
        "704097d0c270e93f0ce5fa91049bb0aa2f38ccfc4bdc38840176abbb98337c0c",
    ],
    [
        "5KKBRfgTgATU5SmF2uy7ewi7BbDDJCtmf3x9CeYziF14uj8YHMM",
        "GPH7pUa1fp4NtGaRDmZF6TeanHw7zELUp1eWxZasRE3zY4xYKdbhV",
        "114aba4ab84ea225bbf4b60aaf6d467d3b206ff8a94d531a5a6031ad90c874dd",
    ],
    [
        "5Jg7muALcVxncN32LyGMDK8zut2b1Sw3VJA1xjZE5ght7DRM9ac",
        "GPH5Vj6uR2iKmrB2DcFyqNzperycD3a32BBYkefzKYCHoGnXemwWS",
        "60928672da8e9a7dc0f783f2bf8aaf1b206b9bbd85f0a61b638e0b99f5f8ea56",
    ],
]


not_enough_padding = [
    {
        "from": "GPH6APYcWtrWXBhcrjPEhPz41bc98NxjnvufVVnRH1M8sjwtvFacz",
        "message_bts": "0b93e05a3b017d00ee16dfea0c1a9d64",
        "message": "#D2BAH3MLo3eMbJh9nR5jy53KXf22b55fQpNLXoGD4bqkE3EkiZirwL8GWsaFJ6g1RDzgRXiYXuNFwCyDddHzuL1SxQrkhcLU2k4tfkcKx1apw8mfzCCJ699LXJxnTgsZd",
        "nonce": "7675159740645758991",
        "plain": "abcdefghijÛ",
        "to": "GPH7Ge953jTDzHKxFAzy19uhJtXxw8CbBM938hkSKWE3yXfRjLV57",
        "wif": "5KR8jzysz2kbYy3TkL3x6NRxfNXwQUWyeVAF5ZagxdqKMawGgXG",
    }
]


class Testcases(unittest.TestCase):
    def test_padding(self):
        for length in range(0, 255):
            s = bytes(length * chr(length), "utf-8")
            padded = _pad(s, 16)
            self.assertEqual(s, _unpad(padded, 16))

    def test_decrypt_bts(self):
        for memo in test_cases:
            dec = decode_memo_bts(
                PrivateKey(memo["wif"]),
                PublicKey(memo["to"], prefix="GPH"),
                memo["nonce"],
                memo["message_bts"],
            )
            self.assertEqual(memo["plain"], dec)

    def test_encrypt_bts(self):
        for memo in test_cases:
            enc = encode_memo_bts(
                PrivateKey(memo["wif"]),
                PublicKey(memo["to"], prefix="GPH"),
                memo["nonce"],
                memo["plain"],
            )
            self.assertEqual(memo["message_bts"], enc)

    def test_decrypt(self):
        for memo in test_cases:
            dec = decode_memo(PrivateKey(memo["wif"]), memo["message"])
            self.assertEqual(memo["plain"], dec[1:])

    def test_encrypt(self):
        for memo in test_cases:
            enc = encode_memo(
                PrivateKey(memo["wif"]),
                PublicKey(memo["to"], prefix="GPH"),
                memo["nonce"],
                memo["plain"],
                prefix="GPH",
            )
            self.assertEqual(memo["message"], enc)

    def test_encrypt_decrypt(self):
        base58 = "#HU6pdQ4Hh8cFrDVooekRPVZu4BdrhAe9RxrWrei2CwfAApAPdM4PT5mSV9cV3tTuWKotYQF6suyM4JHFBZz4pcwyezPzuZ2na7uwhRcLqFoxprno9kWoHiS766vPUKqGX"
        text = "#爱"
        nonce = "1462976530069648"
        wif = str(PasswordKey("", "", role="", prefix="STM").get_private_key())
        private_key = PrivateKey(wif=wif, prefix="STM")
        public_key = private_key.pubkey
        cypertext = encode_memo(private_key, public_key, nonce, text[1:], prefix="STM")
        self.assertEqual(cypertext, base58)
        plaintext = decode_memo(private_key, cypertext)
        self.assertEqual(plaintext, text)

    def test_shared_secret(self):
        for s in test_shared_secrets:
            priv = PrivateKey(s[0])
            pub = PublicKey(s[1], prefix="GPH")
            shared_secret = get_shared_secret(priv, pub)
            self.assertEqual(s[2], shared_secret)

    def test_shared_secrets_equal(self):
        wifs = cycle([x[0] for x in test_shared_secrets])

        for i in range(len(test_shared_secrets)):
            sender_private_key = PrivateKey(next(wifs))
            sender_public_key = sender_private_key.pubkey
            receiver_private_key = PrivateKey(next(wifs))
            receiver_public_key = receiver_private_key.pubkey

            self.assertEqual(
                get_shared_secret(sender_private_key, receiver_public_key),
                get_shared_secret(receiver_private_key, sender_public_key),
            )

    def test_decrypt_bugged_padding_bts(self):
        for memo in not_enough_padding:
            dec = decode_memo_bts(
                PrivateKey(memo["wif"]),
                PublicKey(memo["to"], prefix="GPH"),
                memo["nonce"],
                memo["message_bts"],
            )
            self.assertEqual(memo["plain"], dec)

    def test_decrypt_bugged_padding(self):
        for memo in not_enough_padding:
            dec = decode_memo(PrivateKey(memo["wif"]), memo["message"])
            self.assertEqual(memo["plain"], dec[1:])



# -*- coding: utf-8 -*-


def getBlockParams(ws):
    """Auxiliary method to obtain ``ref_block_num`` and
    ``ref_block_prefix``. Requires a websocket connection to a
    witness node!
    """
    raise DeprecationWarning(
        "This method shouldn't be called anymore. It is part of transactionbuilder now"
    )



# -*- coding: utf-8 -*-
import unittest

from parameterized import parameterized

from nectar import Hive
from nectar.asset import Asset
from nectar.exceptions import AssetDoesNotExistsException
from nectar.instance import set_shared_blockchain_instance

from .nodes import get_hive_nodes


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """
        Initialize class-level Hive instances for tests and register the shared blockchain instance.

        Creates two Hive clients on the test class:
        - cls.bts: uses nodes from get_hive_nodes(), with broadcasting disabled and retries set to 10.
        - cls.hiveio: connected to "https://api.hive.blog", with broadcasting disabled and retries set to 10.

        Registers cls.bts as the global shared blockchain instance via set_shared_blockchain_instance, which is a global side effect relied on by tests.
        """
        cls.bts = Hive(node=get_hive_nodes(), nobroadcast=True, num_retries=10)
        cls.hiveio = Hive(node="https://api.hive.blog", nobroadcast=True, num_retries=10)
        set_shared_blockchain_instance(cls.bts)

    @parameterized.expand(
        [
            ("normal"),
            ("hiveio"),
        ]
    )
    def test_assert(self, node_param):
        if node_param == "normal":
            hv = self.bts
        else:
            hv = self.hiveio
        with self.assertRaises(AssetDoesNotExistsException):
            Asset("FOObarNonExisting", full=False, blockchain_instance=hv)

    @parameterized.expand(
        [
            ("normal", "HBD", "HBD", 3, "@@000000013"),
            ("normal", "HIVE", "HIVE", 3, "@@000000021"),
            ("normal", "VESTS", "VESTS", 6, "@@000000037"),
            ("normal", "@@000000013", "HBD", 3, "@@000000013"),
            ("normal", "@@000000021", "HIVE", 3, "@@000000021"),
            ("normal", "@@000000037", "VESTS", 6, "@@000000037"),
        ]
    )
    def test_properties(self, node_param, data, symbol_str, precision, asset_str):
        if node_param == "normal":
            hv = self.bts
        else:
            hv = self.hiveio
        asset = Asset(data, full=False, blockchain_instance=hv)
        self.assertEqual(asset.symbol, symbol_str)
        self.assertEqual(asset.precision, precision)
        self.assertEqual(asset.asset, asset_str)

    def test_assert_equal(self):
        hv = self.bts
        asset1 = Asset("HBD", full=False, blockchain_instance=hv)
        asset2 = Asset("HBD", full=False, blockchain_instance=hv)
        self.assertTrue(asset1 == asset2)
        self.assertTrue(asset1 == "HBD")
        self.assertTrue(asset2 == "HBD")
        asset3 = Asset("HIVE", full=False, blockchain_instance=hv)
        self.assertTrue(asset1 != asset3)
        self.assertTrue(asset3 != "HBD")
        self.assertTrue(asset1 != "HIVE")

        a = {"asset": "@@000000021", "precision": 3, "id": "HIVE", "symbol": "HIVE"}
        b = {"asset": "@@000000021", "precision": 3, "id": "@@000000021", "symbol": "HIVE"}
        self.assertTrue(Asset(a, blockchain_instance=hv) == Asset(b, blockchain_instance=hv))

    """
    # Mocker comes from pytest-mock, providing an easy way to have patched objects
    # for the life of the test.
    def test_calls(mocker):
        asset = Asset("USD", lazy=True, blockchain_instance=Hive(offline=True))
        method = mocker.patch.object(Asset, 'get_call_orders')
        asset.calls
        method.assert_called_with(10)
    """



# -*- coding: utf-8 -*-
import json
from datetime import date, datetime

from nectar.instance import shared_blockchain_instance
from nectarapi.exceptions import ApiNotSupported

from .blockchainobject import BlockchainObject
from .exceptions import BlockDoesNotExistsException
from .utils import formatTimeString


class Block(BlockchainObject):
    """Read a single block from the chain

    :param int block: block number
    :param Hive blockchain_instance: Hive
        instance
    :param bool lazy: Use lazy loading
    :param bool only_ops: Includes only operations, when set to True (default: False)
    :param bool only_virtual_ops: Includes only virtual operations (default: False)

    Instances of this class are dictionaries that come with additional
    methods (see below) that allow dealing with a block and its
    corresponding functions.

    When only_virtual_ops is set to True, only_ops is always set to True.

    In addition to the block data, the block number is stored as self["id"] or self.identifier.

    .. code-block:: python

        >>> from nectar.block import Block
        >>> block = Block(1)
        >>> print(block)
        <Block 1>

    .. note:: This class comes with its own caching function to reduce the
              load on the API server. Instances of this class can be
              refreshed with ``Account.refresh()``.

    """

    def __init__(
        self,
        block,
        only_ops=False,
        only_virtual_ops=False,
        full=True,
        lazy=False,
        blockchain_instance=None,
        **kwargs,
    ):
        """
        Initialize a Block object representing a single blockchain block.

        block may be an integer (block number), a float (will be converted to int), or a dict containing block data (which will be parsed). Controls:
        - only_ops: load only operations from the block.
        - only_virtual_ops: load only virtual operations.
        - full: if True, populate full block data; if False, keep a minimal representation.
        - lazy: if True, defer fetching full data until needed.

        If no identifier is present after initialization, the block's identifier is set to its numeric block number.
        """
        self.full = full
        self.lazy = lazy
        self.only_ops = only_ops
        self.only_virtual_ops = only_virtual_ops
        if isinstance(block, float):
            block = int(block)
        elif isinstance(block, dict):
            block = self._parse_json_data(block)
        super(Block, self).__init__(
            block, lazy=lazy, full=full, blockchain_instance=blockchain_instance, **kwargs
        )
        if self.identifier is None:
            self.identifier = self.block_num

    def _parse_json_data(self, block):
        parse_times = [
            "timestamp",
        ]
        for p in parse_times:
            if p in block and isinstance(block.get(p), str):
                block[p] = formatTimeString(block.get(p, "1970-01-01T00:00:00"))
        if "transactions" in block:
            for i in range(len(block["transactions"])):
                if "expiration" in block["transactions"][i] and isinstance(
                    block["transactions"][i]["expiration"], str
                ):
                    block["transactions"][i]["expiration"] = formatTimeString(
                        block["transactions"][i]["expiration"]
                    )
        elif "operations" in block:
            for i in range(len(block["operations"])):
                if "timestamp" in block["operations"][i] and isinstance(
                    block["operations"][i]["timestamp"], str
                ):
                    block["operations"][i]["timestamp"] = formatTimeString(
                        block["operations"][i]["timestamp"]
                    )
        return block

    def json(self):
        output = self.copy()
        parse_times = [
            "timestamp",
        ]
        for p in parse_times:
            if p in output:
                p_date = output.get(p, datetime(1970, 1, 1, 0, 0))
                if isinstance(p_date, (datetime, date)):
                    output[p] = formatTimeString(p_date)
                else:
                    output[p] = p_date

        if "transactions" in output:
            for i in range(len(output["transactions"])):
                if "expiration" in output["transactions"][i] and isinstance(
                    output["transactions"][i]["expiration"], (datetime, date)
                ):
                    output["transactions"][i]["expiration"] = formatTimeString(
                        output["transactions"][i]["expiration"]
                    )
        elif "operations" in output:
            for i in range(len(output["operations"])):
                if "timestamp" in output["operations"][i] and isinstance(
                    output["operations"][i]["timestamp"], (datetime, date)
                ):
                    output["operations"][i]["timestamp"] = formatTimeString(
                        output["operations"][i]["timestamp"]
                    )

        ret = json.loads(str(json.dumps(output)))
        output = self._parse_json_data(output)
        return ret

    def refresh(self):
        """Even though blocks never change, you freshly obtain its contents
        from an API with this method
        """
        if self.identifier is None:
            return
        if not self.blockchain.is_connected():
            return
        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        if self.only_ops or self.only_virtual_ops:
            if self.blockchain.rpc.get_use_appbase():
                try:
                    ops_ops = self.blockchain.rpc.get_ops_in_block(
                        {"block_num": self.identifier, "only_virtual": self.only_virtual_ops},
                        api="account_history",
                    )
                    if ops_ops is None:
                        ops = None
                    else:
                        ops = ops_ops["ops"]
                except ApiNotSupported:
                    ops = self.blockchain.rpc.get_ops_in_block(
                        self.identifier, self.only_virtual_ops, api="condenser"
                    )
            else:
                ops = self.blockchain.rpc.get_ops_in_block(self.identifier, self.only_virtual_ops)
            if bool(ops):
                block = {
                    "block": ops[0]["block"],
                    "timestamp": ops[0]["timestamp"],
                    "operations": ops,
                }
            else:
                block = {
                    "block": self.identifier,
                    "timestamp": "1970-01-01T00:00:00",
                    "operations": [],
                }
        else:
            if self.blockchain.rpc.get_use_appbase():
                try:
                    block = self.blockchain.rpc.get_block(
                        {"block_num": self.identifier}, api="block"
                    )
                    if block and "block" in block:
                        block = block["block"]
                except ApiNotSupported:
                    block = self.blockchain.rpc.get_block(self.identifier, api="condenser")
            else:
                block = self.blockchain.rpc.get_block(self.identifier)
        if not block:
            message = f"Block {self.identifier} does not exist or is not available from {self.blockchain.rpc.url}"
            raise BlockDoesNotExistsException(message)
        block = self._parse_json_data(block)
        super(Block, self).__init__(
            block, lazy=self.lazy, full=self.full, blockchain_instance=self.blockchain
        )

    @property
    def block_num(self):
        """Returns the block number"""
        if "block_id" in self:
            return int(self["block_id"][:8], base=16)
        elif "block" in self:
            return int(self["block"])
        else:
            return None

    def time(self):
        """Return a datetime instance for the timestamp of this block"""
        return self["timestamp"]

    @property
    def transactions(self):
        """Returns all transactions as list"""
        if self.only_ops or self.only_virtual_ops:
            return list()
        trxs = []
        if "transactions" not in self:
            return []
        trx_id = 0
        for trx in self["transactions"]:
            trx_new = {"transaction_id": self["transaction_ids"][trx_id]}
            trx_new.update(trx.copy())
            trx_new.update({"block_num": self.block_num, "transaction_num": trx_id})
            trxs.append(trx_new)
            trx_id += 1
        return trxs

    @property
    def operations(self):
        """Returns all block operations as list"""
        if self.only_ops or self.only_virtual_ops:
            return self["operations"]
        ops = []
        trxs = []
        if "transactions" in self:
            trxs = self["transactions"]
        for tx in trxs:
            if "operations" not in tx:
                continue
            for op in tx["operations"]:
                # Replace opid by op name
                # op[0] = getOperationNameForId(op[0])
                if isinstance(op, list):
                    ops.append(list(op))
                else:
                    ops.append(op.copy())
        return ops

    @property
    def json_transactions(self):
        """Returns all transactions as list, all dates are strings."""
        if self.only_ops or self.only_virtual_ops:
            return list()
        trxs = []
        if "transactions" not in self:
            return []
        trx_id = 0
        for trx in self["transactions"]:
            trx_new = {"transaction_id": self["transaction_ids"][trx_id]}
            trx_new.update(trx.copy())
            trx_new.update({"block_num": self.block_num, "transaction_num": trx_id})
            if "expiration" in trx:
                p_date = trx.get("expiration", datetime(1970, 1, 1, 0, 0))
                if isinstance(p_date, (datetime, date)):
                    trx_new.update({"expiration": formatTimeString(p_date)})

            trxs.append(trx_new)
            trx_id += 1
        return trxs

    @property
    def json_operations(self):
        """Returns all block operations as list, all dates are strings."""
        if self.only_ops or self.only_virtual_ops:
            return self["operations"]
        ops = []
        for tx in self["transactions"]:
            for op in tx["operations"]:
                if "operations" not in tx:
                    continue
                # Replace opid by op name
                # op[0] = getOperationNameForId(op[0])
                if isinstance(op, list):
                    op_new = list(op)
                else:
                    op_new = op.copy()
                if "timestamp" in op:
                    p_date = op.get("timestamp", datetime(1970, 1, 1, 0, 0))
                    if isinstance(p_date, (datetime, date)):
                        op_new.update({"timestamp": formatTimeString(p_date)})
                ops.append(op_new)
        return ops

    def ops_statistics(self, add_to_ops_stat=None):
        """Returns a statistic with the occurrence of the different operation types"""
        if add_to_ops_stat is None:
            import nectarbase.operationids

            ops_stat = nectarbase.operationids.operations.copy()
            for key in ops_stat:
                if isinstance(key, int):
                    ops_stat[key] = 0
        else:
            ops_stat = add_to_ops_stat.copy()
        for op in self.operations:
            if "op" in op:
                op = op["op"]
            if isinstance(op, dict) and "type" in op:
                op_type = op["type"]
                if len(op_type) > 10 and op_type[len(op_type) - 10 :] == "_operation":
                    op_type = op_type[:-10]
            else:
                op_type = op[0]
            if op_type in ops_stat:
                ops_stat[op_type] += 1
        return ops_stat


class BlockHeader(BlockchainObject):
    """Read a single block header from the chain

    :param int block: block number
    :param Hive blockchain_instance: Hive
        instance
    :param bool lazy: Use lazy loading

    In addition to the block data, the block number is stored as self["id"] or self.identifier.

    .. code-block:: python

        >>> from nectar.block import BlockHeader
        >>> block = BlockHeader(1)
        >>> print(block)
        <BlockHeader 1>

    """

    def __init__(self, block, full=True, lazy=False, blockchain_instance=None, **kwargs):
        """
        Initialize a BlockHeader.

        One-line summary:
            Create a BlockHeader wrapper for a block header, optionally in lazy or full mode.

        Parameters:
            block (int | float | dict): Block number (floats are converted to int) or a header dict.
            full (bool): If True, populate the object with full header data; otherwise keep a minimal representation.
            lazy (bool): If True, delay API fetching until data is accessed.

        Notes:
            If no blockchain_instance is provided, the module's shared blockchain instance is used.
        """
        self.full = full
        self.lazy = lazy
        if isinstance(block, float):
            block = int(block)
        super(BlockHeader, self).__init__(
            block, lazy=lazy, full=full, blockchain_instance=blockchain_instance, **kwargs
        )

    def refresh(self):
        """Even though blocks never change, you freshly obtain its contents
        from an API with this method
        """
        if not self.blockchain.is_connected():
            return None
        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        if self.blockchain.rpc.get_use_appbase():
            block = self.blockchain.rpc.get_block_header(
                {"block_num": self.identifier}, api="block"
            )
            if block is not None and "header" in block:
                block = block["header"]
        else:
            block = self.blockchain.rpc.get_block_header(self.identifier)
        if not block:
            raise BlockDoesNotExistsException(str(self.identifier))
        block = self._parse_json_data(block)
        super(BlockHeader, self).__init__(
            block, lazy=self.lazy, full=self.full, blockchain_instance=self.blockchain
        )

    def time(self):
        """Return a datetime instance for the timestamp of this block"""
        return self["timestamp"]

    @property
    def block_num(self):
        """Returns the block number"""
        return self.identifier

    def _parse_json_data(self, block):
        parse_times = [
            "timestamp",
        ]
        for p in parse_times:
            if p in block and isinstance(block.get(p), str):
                block[p] = formatTimeString(block.get(p, "1970-01-01T00:00:00"))
        return block

    def json(self):
        output = self.copy()
        parse_times = [
            "timestamp",
        ]
        for p in parse_times:
            if p in output:
                p_date = output.get(p, datetime(1970, 1, 1, 0, 0))
                if isinstance(p_date, (datetime, date)):
                    output[p] = formatTimeString(p_date)
                else:
                    output[p] = p_date
        return json.loads(str(json.dumps(output)))


class Blocks(list):
    """Obtain a list of blocks

    :param list name_list: list of accounts to fetch
    :param int count: (optional) maximum number of accounts
        to fetch per call, defaults to 100
    :param Hive blockchain_instance: Hive() instance to use when
        accessing RPC
    """

    def __init__(
        self,
        starting_block_num,
        count=1000,
        lazy=False,
        full=True,
        only_ops=False,
        only_virtual_ops=False,
        blockchain_instance=None,
        **kwargs,
    ):
        """
        Initialize a Blocks collection by fetching a contiguous range of blocks from the chain and populating the list with Block objects.

        If a blockchain_instance is provided it is used; otherwise the shared blockchain instance is used. If the chosen instance is not connected, the initializer returns early and the Blocks object remains empty.

        Parameters:
            starting_block_num (int): First block number to retrieve.
            count (int, optional): Number of consecutive blocks to fetch. Defaults to 1000.
            lazy (bool, optional): If True, create Block objects in lazy mode (defer full parsing). Defaults to False.
            full (bool, optional): If True, create Block objects with full data loaded (subject to lazy). Defaults to True.
            only_ops (bool, optional): If True, blocks will contain only regular operations (no block metadata). Defaults to False.
            only_virtual_ops (bool, optional): If True, blocks will contain only virtual operations. Defaults to False.
        """
        self.blockchain = blockchain_instance or shared_blockchain_instance()

        if not self.blockchain.is_connected():
            return
        blocks = []

        self.blockchain.rpc.set_next_node_on_empty_reply(False)

        blocks = self.blockchain.rpc.get_block_range(
            {"starting_block_num": starting_block_num, "count": count}, api="block"
        )["blocks"]

        super(Blocks, self).__init__(
            [
                Block(
                    x,
                    lazy=lazy,
                    full=full,
                    blockchain_instance=self.blockchain,
                )
                for x in blocks
            ]
        )



# -*- coding: utf-8 -*-

from nectar.constants import (
    EXEC_FOLLOW_CUSTOM_OP_SCALE,
    resource_execution_time,
    state_object_size_info,
)
from nectarbase import operations
from nectarbase.objects import Operation
from nectarbase.signedtransactions import Signed_Transaction

from .instance import shared_blockchain_instance


class RC(object):
    def __init__(self, blockchain_instance=None, **kwargs):
        """
        Initialize the RC helper with a blockchain instance.

        If `blockchain_instance` is provided it will be used for RC lookups and broadcasts;
        otherwise the module-wide shared_blockchain_instance() is used. Extra keyword
        arguments are accepted for compatibility but ignored.
        """
        self.blockchain = blockchain_instance or shared_blockchain_instance()

    def get_tx_size(self, op):
        """
        Estimate the serialized size (in bytes) of a signed transaction containing the given operation.

        This constructs a dummy Signed_Transaction using fixed reference fields and a hard-coded private key, signs it on the "HIVE" chain, and returns the length of the resulting serialized transaction in bytes. The value is an estimate useful for RC sizing and does not represent a real broadcastable transaction.

        Parameters:
            op: Operation or dict-like operation payload to include in the transaction.

        Returns:
            int: Number of bytes in the serialized, signed transaction.
        """
        ops = [Operation(op)]
        prefix = "HIVE"
        wif = "5KQwrPbwdL6PhXujxW37FSSQZ1JiwsST4cqQzDeyXtP79zkvFD3"
        ref_block_num = 34294
        ref_block_prefix = 3707022213
        expiration = "2016-04-06T08:29:27"
        tx = Signed_Transaction(
            ref_block_num=ref_block_num,
            ref_block_prefix=ref_block_prefix,
            expiration=expiration,
            operations=ops,
        )
        tx = tx.sign([wif], chain=prefix)
        tx_size = len(bytes(tx))
        return tx_size

    def get_resource_count(
        self,
        tx_size,
        execution_time_count,
        state_bytes_count=0,
        new_account_op_count=0,
        market_op_count=0,
    ):
        """
        Build and return a resource_count mapping for RC cost calculation.

        Parameters:
            tx_size (int): Transaction size in bytes; used for history bytes and for market bytes when applicable.
            execution_time_count (int): Execution time units for the operation.
            state_bytes_count (int, optional): Additional state bytes contributed by the operation (default 0).
            new_account_op_count (int, optional): Number of new-account operations included (default 0).
            market_op_count (int, optional): If > 0, marks the transaction as a market operation and sets market bytes to tx_size (default 0).

        Returns:
            dict: A dictionary containing keys used by the RC pricing engine, including:
                - resource_history_bytes
                - resource_state_bytes
                - resource_new_accounts
                - resource_execution_time
                - resource_market_bytes (only present if market_op_count > 0)
        """
        resource_count = {"resource_history_bytes": tx_size}
        resource_count["resource_state_bytes"] = state_object_size_info[
            "transaction_object_base_size"
        ]
        resource_count["resource_state_bytes"] += (
            state_object_size_info["transaction_object_byte_size"] * tx_size
        )
        resource_count["resource_state_bytes"] += state_bytes_count
        resource_count["resource_new_accounts"] = new_account_op_count
        resource_count["resource_execution_time"] = execution_time_count
        if market_op_count > 0:
            resource_count["resource_market_bytes"] = tx_size
        return resource_count

    def comment_dict(self, comment_dict):
        """Calc RC costs for a comment dict object

        Example for calculating RC costs

        .. code-block:: python

            from nectar.rc import RC
            comment_dict = {
                            "permlink": "test", "author": "thecrazygm",
                            "body": "test", "parent_permlink": "",
                            "parent_author": "", "title": "test",
                            "json_metadata": {"foo": "bar"}
                           }

            rc = RC()
            print(rc.comment_from_dict(comment_dict))

        """
        op = operations.Comment(**comment_dict)
        tx_size = self.get_tx_size(op)
        permlink_length = len(comment_dict["permlink"])
        parent_permlink_length = len(comment_dict["parent_permlink"])
        return self.comment(
            tx_size=tx_size,
            permlink_length=permlink_length,
            parent_permlink_length=parent_permlink_length,
        )

    def comment(self, tx_size=1000, permlink_length=10, parent_permlink_length=10):
        """Calc RC for a comment"""
        state_bytes_count = state_object_size_info["comment_object_base_size"]
        state_bytes_count += (
            state_object_size_info["comment_object_permlink_char_size"] * permlink_length
        )
        state_bytes_count += (
            state_object_size_info["comment_object_parent_permlink_char_size"]
            * parent_permlink_length
        )
        execution_time_count = resource_execution_time["comment_operation_exec_time"]
        resource_count = self.get_resource_count(tx_size, execution_time_count, state_bytes_count)
        return self.blockchain.get_rc_cost(resource_count)

    def vote_dict(self, vote_dict):
        """Calc RC costs for a vote

        Example for calculating RC costs

        .. code-block:: python

            from nectar.rc import RC
            vote_dict = {
                         "voter": "foobara", "author": "foobarc",
                         "permlink": "foobard", "weight": 1000
                        }

            rc = RC()
            print(rc.comment(vote_dict))

        """
        op = operations.Vote(**vote_dict)
        tx_size = self.get_tx_size(op)
        return self.vote(tx_size=tx_size)

    def vote(self, tx_size=210):
        """Calc RC for a vote"""
        state_bytes_count = state_object_size_info["comment_vote_object_base_size"]
        execution_time_count = resource_execution_time["vote_operation_exec_time"]
        resource_count = self.get_resource_count(tx_size, execution_time_count, state_bytes_count)
        return self.blockchain.get_rc_cost(resource_count)

    def transfer_dict(self, transfer_dict):
        """
        Calculate Resource Credit (RC) cost for a transfer operation represented as a dict.

        The input dict must contain the fields required by a Transfer operation (for example: "from", "to", "amount", "memo"). This function builds a Transfer operation, estimates the signed transaction size, marks the operation as a market operation (market_op_count=1), and returns the RC cost computed by the blockchain instance.

        Parameters:
            transfer_dict (dict): Fields for a Transfer operation compatible with operations.Transfer.

        Returns:
            dict: RC cost structure as returned by the blockchain's get_rc_cost.
        """
        market_op_count = 1
        op = operations.Transfer(**transfer_dict)
        tx_size = self.get_tx_size(op)
        return self.transfer(tx_size=tx_size, market_op_count=market_op_count)

    def transfer(self, tx_size=290, market_op_count=1):
        """Calc RC of a transfer"""
        execution_time_count = resource_execution_time["transfer_operation_exec_time"]
        resource_count = self.get_resource_count(
            tx_size, execution_time_count, market_op_count=market_op_count
        )
        return self.blockchain.get_rc_cost(resource_count)

    def custom_json_dict(self, custom_json_dict):
        """Calc RC costs for a custom_json

        Example for calculating RC costs

        .. code-block:: python

            from nectar.rc import RC
            from collections import OrderedDict
            custom_json_dict = {
                                 "json": [
                                          "reblog", OrderedDict([("account", "xeroc"), ("author", "chainsquad"),
                                                                 ("permlink", "streemian-com-to-open-its-doors-and-offer-a-20-discount")
                                                                ])
                                         ],
                                 "required_auths": [],
                                 "required_posting_auths": ["xeroc"],
                                 "id": "follow"
                                }

            rc = RC()
            print(rc.comment(custom_json_dict))

        """
        op = operations.Custom_json(**custom_json_dict)
        tx_size = self.get_tx_size(op)
        follow_id = custom_json_dict["id"] == "follow"
        return self.custom_json(tx_size=tx_size, follow_id=follow_id)

    def custom_json(self, tx_size=444, follow_id=False):
        execution_time_count = resource_execution_time["custom_json_operation_exec_time"]
        if follow_id:
            execution_time_count *= EXEC_FOLLOW_CUSTOM_OP_SCALE
        resource_count = self.get_resource_count(tx_size, execution_time_count)
        return self.blockchain.get_rc_cost(resource_count)

    def account_update_dict(self, account_update_dict):
        """Calc RC costs for account update"""
        op = operations.Account_update(**account_update_dict)
        tx_size = self.get_tx_size(op)
        execution_time_count = resource_execution_time["account_update_operation_exec_time"]
        resource_count = self.get_resource_count(tx_size, execution_time_count)
        return self.blockchain.get_rc_cost(resource_count)

    def claim_account(self, tx_size=300):
        """Claim account"""
        execution_time_count = resource_execution_time["claim_account_operation_exec_time"]
        resource_count = self.get_resource_count(
            tx_size, execution_time_count, new_account_op_count=1
        )
        return self.blockchain.get_rc_cost(resource_count)

    def get_authority_byte_count(self, auth):
        return (
            state_object_size_info["authority_base_size"]
            + state_object_size_info["authority_account_member_size"] * len(auth["account_auths"])
            + state_object_size_info["authority_key_member_size"] * len(auth["key_auths"])
        )

    def account_create_dict(self, account_create_dict):
        """Calc RC costs for account create"""
        op = operations.Account_create(**account_create_dict)
        state_bytes_count = state_object_size_info["account_object_base_size"]
        state_bytes_count += state_object_size_info["account_authority_object_base_size"]
        state_bytes_count += self.get_authority_byte_count(account_create_dict["owner"])
        state_bytes_count += self.get_authority_byte_count(account_create_dict["active"])
        state_bytes_count += self.get_authority_byte_count(account_create_dict["posting"])
        tx_size = self.get_tx_size(op)
        execution_time_count = resource_execution_time["account_update_operation_exec_time"]
        resource_count = self.get_resource_count(tx_size, execution_time_count, state_bytes_count)
        return self.blockchain.get_rc_cost(resource_count)

    def create_claimed_account_dict(self, create_claimed_account_dict):
        """Calc RC costs for claimed account create"""
        op = operations.Create_claimed_account(**create_claimed_account_dict)
        state_bytes_count = state_object_size_info["account_object_base_size"]
        state_bytes_count += state_object_size_info["account_authority_object_base_size"]
        state_bytes_count += self.get_authority_byte_count(create_claimed_account_dict["owner"])
        state_bytes_count += self.get_authority_byte_count(create_claimed_account_dict["active"])
        state_bytes_count += self.get_authority_byte_count(create_claimed_account_dict["posting"])
        tx_size = self.get_tx_size(op)
        execution_time_count = resource_execution_time["account_update_operation_exec_time"]
        resource_count = self.get_resource_count(tx_size, execution_time_count, state_bytes_count)
        return self.blockchain.get_rc_cost(resource_count)

    def set_slot_delegator(self, from_pool, to_account, to_slot, signer):
        """Set a slot to receive RC from a pool

        :param str from_pool: Pool to set the slot to
        :param str to_account: Account on which we want to update the slot
        :param int to_slot: slot we want to set
        :param str signer: Account who broadcast this
        """
        json_body = [
            "set_slot_delegator",
            {
                "from_pool": from_pool,
                "to_account": to_account,
                "to_slot": to_slot,
                "signer": signer,
            },
        ]
        return self.blockchain.custom_json("rc", json_body, required_auths=[signer])

    def delegate_from_pool(self, from_pool, to_account, max_rc):
        """Set a slot to receive RC from a pool

        :param str from_pool: Pool to set the slot to
        :param str to_account: Account on which we want to update the slot
        :param int max_rc: max rc to delegate
        """
        json_body = [
            "delegate_drc_from_pool",
            {
                "from_pool": from_pool,
                "to_account": to_account,
                "asset_symbol": {"nai": "@@000000037", "decimals": 6},
                "drc_max_mana": max_rc,
            },
        ]
        return self.blockchain.custom_json("rc", json_body, required_auths=[from_pool])

    def delegate_to_pool(self, username, to_pool, rc):
        """Set a slot to receive RC from a pool

        :param str username: user delegating rc to the pool
        :param str to_pool: Pool to delegate to
        :param str rc: rc to delegate
        """
        json_body = [
            "delegate_to_pool",
            {
                "from_account": username,
                "to_pool": to_pool,
                "amount": {"symbol": "VESTS", "amount": rc, "precision": 6, "nai": "@@000000037"},
            },
        ]
        return self.blockchain.custom_json("rc", json_body, required_auths=[username])



# -*- coding: utf-8 -*-
#: Operation ids

# https://gitlab.syncad.com/hive/hive/-/blob/master/libraries/protocol/include/hive/protocol/operations.hpp
ops = [
    "vote",  # 0
    "comment",  # 1
    "transfer",  # 2
    "transfer_to_vesting",  # 3
    "withdraw_vesting",  # 4
    "limit_order_create",  # 5
    "limit_order_cancel",  # 6
    "feed_publish",  # 7
    "convert",  # 8
    "account_create",  # 9
    "account_update",  # 10
    "witness_update",  # 11
    "account_witness_vote",  # 12
    "account_witness_proxy",  # 13
    "pow",  # 14
    "custom",  # 15
    "report_over_production",  # 16
    "delete_comment",  # 17
    "custom_json",  # 18
    "comment_options",  # 19
    "set_withdraw_vesting_route",  # 20
    "limit_order_create2",  # 21
    "claim_account",  # 22
    "create_claimed_account",  # 23
    "request_account_recovery",  # 24
    "recover_account",  # 25
    "change_recovery_account",  # 26
    "escrow_transfer",  # 27
    "escrow_dispute",  # 28
    "escrow_release",  # 29
    "pow2",  # 30
    "escrow_approve",  # 31
    "transfer_to_savings",  # 32
    "transfer_from_savings",  # 33
    "cancel_transfer_from_savings",  # 34
    "custom_binary",  # 35
    "decline_voting_rights",  # 36
    "reset_account",  # 37
    "set_reset_account",  # 38
    "claim_reward_balance",  # 39
    "delegate_vesting_shares",  # 40
    "account_create_with_delegation",  # 41
    "witness_set_properties",  # 42
    "account_update2",  # 43
    "create_proposal",  # 44
    "update_proposal_votes",  # 45
    "remove_proposal",  # 46
    "update_proposal",  # 47
    "collateralized_convert",  # 48
    "recurrent_transfer",  # 49
    # virtual operations below this point
    "fill_convert_request",  # last_regular + 1
    "author_reward",  # last_regular + 2
    "curation_reward",  # last_regular + 3
    "comment_reward",  # last_regular + 4
    "liquidity_reward",  # last_regular + 5
    "interest",  # last_regular + 6
    "fill_vesting_withdraw",  # last_regular + 7
    "fill_order",  # last_regular + 8
    "shutdown_witness",  # last_regular + 9
    "fill_transfer_from_savings",  # last_regular + 10
    "hardfork",  # last_regular + 11
    "comment_payout_update",  # last_regular + 12
    "return_vesting_delegation",  # last_regular + 13
    "comment_benefactor_reward",  # last_regular + 14
    "producer_reward",  # last_regular + 15
    "clear_null_account_balance",  # last_regular + 16
    "proposal_pay",  # last_regular + 17
    "sps_fund",  # last_regular + 18
    "hardfork_hive",  # last_regular + 19
    "hardfork_hive_restore",  # last_regular + 20
    "delayed_voting",  # last_regular + 21
    "consolidate_treasury_balance",  # last_regular + 22
    "effective_comment_vote",  # last_regular + 23
    "ineffective_delete_comment",  # last_regular + 24
    "sps_convert",  # last_regular + 25
    "expired_account_notification",  # last_regular + 26
    "changed_recovery_account",  # last_regular + 27
    "transfer_to_vesting_completed",  # last_regular + 28
    "pow_reward",  # last_regular + 29
    "vesting_shares_split",  # last_regular + 30
    "account_created",  # last_regular + 31
    "fill_collateralized_convert_request",  # last_regular + 32
    "system_warning",  # last_regular + 33,
    "fill_recurrent_transfer",  # last_regular + 34
    "failed_recurrent_transfer",  # last_regular + 35
]

operations = {o: ops.index(o) for o in ops}


def getOperationNameForId(i):
    """Convert an operation id into the corresponding string"""
    for key in operations:
        if int(operations[key]) is int(i):
            return key
    return "Unknown Operation ID %d" % i



# -*- coding: utf-8 -*-
from decimal import ROUND_DOWN, Decimal

from nectar.asset import Asset
from nectar.instance import shared_blockchain_instance


def check_asset(other, self, hv):
    """
    Assert that two asset representations refer to the same asset.

    If both `other` and `self` are dicts containing an "asset" key, each asset id is wrapped in an Asset using the provided blockchain instance and compared for equality. Otherwise the two values are compared directly. Raises AssertionError if the values do not match.
    """
    if isinstance(other, dict) and "asset" in other and isinstance(self, dict) and "asset" in self:
        if not Asset(other["asset"], blockchain_instance=hv) == Asset(
            self["asset"], blockchain_instance=hv
        ):
            raise AssertionError()
    else:
        if not other == self:
            raise AssertionError()


def quantize(amount, precision):
    # make sure amount is decimal and has the asset precision
    amount = Decimal(amount)
    places = Decimal(10) ** (-precision)
    return amount.quantize(places, rounding=ROUND_DOWN)


class Amount(dict):
    """This class deals with Amounts of any asset to simplify dealing with the tuple::

        (amount, asset)

    :param list args: Allows to deal with different representations of an amount
    :param float amount: Let's create an instance with a specific amount
    :param str asset: Let's you create an instance with a specific asset (symbol)
    :param boolean fixed_point_arithmetic: when set to True, all operations are fixed
        point operations and the amount is always be rounded down to the precision
    :param Blockchain blockchain_instance: Blockchain instance
    :returns: All data required to represent an Amount/Asset
    :rtype: dict
    :raises ValueError: if the data provided is not recognized

    Way to obtain a proper instance:

        * ``args`` can be a string, e.g.:  "1 HBD"
        * ``args`` can be a dictionary containing ``amount`` and ``asset_id``
        * ``args`` can be a dictionary containing ``amount`` and ``asset``
        * ``args`` can be a list of a ``float`` and ``str`` (symbol)
        * ``args`` can be a list of a ``float`` and a :class:`nectar.asset.Asset`
        * ``amount`` and ``asset`` are defined manually

    An instance is a dictionary and comes with the following keys:

        * ``amount`` (float)
        * ``symbol`` (str)
        * ``asset`` (instance of :class:`nectar.asset.Asset`)

    Instances of this class can be used in regular mathematical expressions
    (``+-*/%``) such as:

    .. testcode::

        from nectar.amount import Amount
        from nectar.asset import Asset
        a = Amount("1 HIVE")
        b = Amount(1, "HIVE")
        c = Amount("20", Asset("HIVE"))
        a + b
        a * 2
        a += b
        a /= 2.0

    .. testoutput::

        2.000 HIVE
        2.000 HIVE

    """

    def __init__(
        self,
        amount,
        asset=None,
        fixed_point_arithmetic=False,
        new_appbase_format=True,
        blockchain_instance=None,
        **kwargs,
    ):
        """
        Initialize an Amount object representing a quantity of a specific blockchain asset.

        The constructor accepts many input formats and normalizes them into internal keys:
        - amount may be another Amount, a three-element list [amount, precision, asset],
          a new appbase-format dict with keys ("amount", "nai", "precision"), a legacy dict
          with ("amount", "asset_id") or ("amount", "asset"), a string like "1.000 HIVE",
          or a numeric value (int, float, Decimal) paired with an `asset` argument.
        - asset may be an Asset instance, an asset dict, or a symbol string; when omitted,
          the asset will be inferred from the provided `amount` representation.

        After parsing, the instance stores:
        - "amount" as a Decimal (or quantized Decimal when fixed-point mode is enabled),
        - "symbol" as the asset symbol,
        - "asset" as an Asset-like object.

        Parameters:
            amount: Various accepted formats (see description) representing the quantity.
            asset: Asset instance, asset dict, or asset symbol string used when `amount`
                is a bare numeric value or when explicit asset resolution is required.
            fixed_point_arithmetic (bool): When True, the numeric amount is quantized
                to the asset's precision using floor rounding.
            new_appbase_format (bool): Indicates whether to prefer the new appbase JSON
                format when producing serialized representations.

        Raises:
            ValueError: If `amount` and `asset` do not match any supported input format.
        """
        self["asset"] = {}
        self.new_appbase_format = new_appbase_format
        self.fixed_point_arithmetic = fixed_point_arithmetic

        self.blockchain = blockchain_instance or shared_blockchain_instance()

        if amount and asset is None and isinstance(amount, Amount):
            # Copy Asset object
            self["amount"] = amount["amount"]
            self["symbol"] = amount["symbol"]
            self["asset"] = amount["asset"]

        elif amount and asset is None and isinstance(amount, list) and len(amount) == 3:
            # Copy Asset object
            self["amount"] = Decimal(amount[0]) / Decimal(10 ** amount[1])
            self["asset"] = Asset(amount[2], blockchain_instance=self.blockchain)
            self["symbol"] = self["asset"]["symbol"]

        elif (
            amount
            and asset is None
            and isinstance(amount, dict)
            and "amount" in amount
            and "nai" in amount
            and "precision" in amount
        ):
            # Copy Asset object
            self.new_appbase_format = True
            self["amount"] = Decimal(amount["amount"]) / Decimal(10 ** amount["precision"])
            self["asset"] = Asset(amount["nai"], blockchain_instance=self.blockchain)
            self["symbol"] = self["asset"]["symbol"]

        elif amount is not None and asset is None and isinstance(amount, str):
            self["amount"], self["symbol"] = amount.split(" ")
            self["asset"] = Asset(self["symbol"], blockchain_instance=self.blockchain)

        elif (
            amount
            and asset is None
            and isinstance(amount, dict)
            and "amount" in amount
            and "asset_id" in amount
        ):
            self["asset"] = Asset(amount["asset_id"], blockchain_instance=self.blockchain)
            self["symbol"] = self["asset"]["symbol"]
            self["amount"] = Decimal(amount["amount"]) / Decimal(10 ** self["asset"]["precision"])

        elif (
            amount
            and asset is None
            and isinstance(amount, dict)
            and "amount" in amount
            and "asset" in amount
        ):
            self["asset"] = Asset(amount["asset"], blockchain_instance=self.blockchain)
            self["symbol"] = self["asset"]["symbol"]
            self["amount"] = Decimal(amount["amount"]) / Decimal(10 ** self["asset"]["precision"])

        elif isinstance(amount, (float)) and asset and isinstance(asset, Asset):
            self["amount"] = str(amount)
            self["asset"] = asset
            self["symbol"] = self["asset"]["symbol"]

        elif isinstance(amount, (int, Decimal)) and asset and isinstance(asset, Asset):
            self["amount"] = amount
            self["asset"] = asset
            self["symbol"] = self["asset"]["symbol"]

        elif isinstance(amount, (float)) and asset and isinstance(asset, dict):
            self["amount"] = str(amount)
            self["asset"] = asset
            self["symbol"] = self["asset"]["symbol"]

        elif isinstance(amount, (int, Decimal)) and asset and isinstance(asset, dict):
            self["amount"] = amount
            self["asset"] = asset
            self["symbol"] = self["asset"]["symbol"]

        elif isinstance(amount, (float)) and asset and isinstance(asset, str):
            self["amount"] = str(amount)
            self["asset"] = Asset(asset, blockchain_instance=self.blockchain)
            self["symbol"] = asset

        elif isinstance(amount, (int, Decimal)) and asset and isinstance(asset, str):
            self["amount"] = amount
            self["asset"] = Asset(asset, blockchain_instance=self.blockchain)
            self["symbol"] = asset
        elif amount and asset and isinstance(asset, Asset):
            self["amount"] = amount
            self["symbol"] = asset["symbol"]
            self["asset"] = asset
        elif amount and asset and isinstance(asset, str):
            self["amount"] = amount
            self["asset"] = Asset(asset, blockchain_instance=self.blockchain)
            self["symbol"] = self["asset"]["symbol"]
        else:
            raise ValueError
        if self.fixed_point_arithmetic:
            self["amount"] = quantize(self["amount"], self["asset"]["precision"])
        else:
            self["amount"] = Decimal(self["amount"])

    def copy(self):
        """Copy the instance and make sure not to use a reference"""
        return Amount(
            amount=self["amount"],
            asset=self["asset"].copy(),
            new_appbase_format=self.new_appbase_format,
            fixed_point_arithmetic=self.fixed_point_arithmetic,
            blockchain_instance=self.blockchain,
        )

    @property
    def amount(self):
        """Returns the amount as float"""
        return float(self["amount"])

    @property
    def amount_decimal(self):
        """Returns the amount as decimal"""
        return self["amount"]

    @property
    def symbol(self):
        """Returns the symbol of the asset"""
        return self["symbol"]

    def tuple(self):
        return float(self), self.symbol

    @property
    def asset(self):
        """
        Return the Asset object for this Amount, constructing it lazily if missing.

        If the internal 'asset' entry is falsy, this creates a nectar.asset.Asset using the stored symbol
        and this Amount's blockchain instance, stores it in 'asset', and returns it. Always returns an
        Asset instance.
        """
        if not isinstance(self["asset"], Asset):
            self["asset"] = Asset(self["symbol"], blockchain_instance=self.blockchain)
        return self["asset"]

    def json(self):
        if self.blockchain.is_connected() and self.blockchain.rpc.get_use_appbase():
            if self.new_appbase_format:
                return {
                    "amount": str(int(self)),
                    "nai": self["asset"]["asset"],
                    "precision": self["asset"]["precision"],
                }
            else:
                return [str(int(self)), self["asset"]["precision"], self["asset"]["asset"]]
        else:
            return str(self)

    def __str__(self):
        amount = quantize(self["amount"], self["asset"]["precision"])
        symbol = self["symbol"]
        return "{:.{prec}f} {}".format(amount, symbol, prec=self["asset"]["precision"])

    def __float__(self):
        if self.fixed_point_arithmetic:
            return float(quantize(self["amount"], self["asset"]["precision"]))
        else:
            return float(self["amount"])

    def __int__(self):
        amount = quantize(self["amount"], self["asset"]["precision"])
        return int(amount * 10 ** self["asset"]["precision"])

    def __add__(self, other):
        a = self.copy()
        if isinstance(other, Amount):
            check_asset(other["asset"], self["asset"], self.blockchain)
            a["amount"] += other["amount"]
        else:
            a["amount"] += Decimal(other)
        if self.fixed_point_arithmetic:
            a["amount"] = quantize(a["amount"], self["asset"]["precision"])
        return a

    def __sub__(self, other):
        a = self.copy()
        if isinstance(other, Amount):
            check_asset(other["asset"], self["asset"], self.blockchain)
            a["amount"] -= other["amount"]
        else:
            a["amount"] -= Decimal(other)
        if self.fixed_point_arithmetic:
            a["amount"] = quantize(a["amount"], self["asset"]["precision"])
        return a

    def __mul__(self, other):
        from .price import Price

        a = self.copy()
        if isinstance(other, Amount):
            check_asset(other["asset"], self["asset"], self.blockchain)
            a["amount"] *= other["amount"]
        elif isinstance(other, Price):
            if not self["asset"] == other["quote"]["asset"]:
                raise AssertionError()
            a = self.copy() * other["price"]
            a["asset"] = other["base"]["asset"].copy()
            a["symbol"] = other["base"]["asset"]["symbol"]
        else:
            a["amount"] *= Decimal(other)
        if self.fixed_point_arithmetic:
            a["amount"] = quantize(a["amount"], self["asset"]["precision"])
        return a

    def __floordiv__(self, other):
        a = self.copy()
        if isinstance(other, Amount):
            from .price import Price

            check_asset(other["asset"], self["asset"], self.blockchain)
            return Price(self, other, blockchain_instance=self.blockchain)
        else:
            a["amount"] //= Decimal(other)
        if self.fixed_point_arithmetic:
            a["amount"] = quantize(a["amount"], self["asset"]["precision"])
        return a

    def __div__(self, other):
        from .price import Price

        a = self.copy()
        if isinstance(other, Amount):
            check_asset(other["asset"], self["asset"], self.blockchain)
            return Price(self, other, blockchain_instance=self.blockchain)
        elif isinstance(other, Price):
            if not self["asset"] == other["base"]["asset"]:
                raise AssertionError()
            a = self.copy()
            a["amount"] = a["amount"] / other["price"]
            a["asset"] = other["quote"]["asset"].copy()
            a["symbol"] = other["quote"]["asset"]["symbol"]
        else:
            a["amount"] /= Decimal(other)
        if self.fixed_point_arithmetic:
            a["amount"] = quantize(a["amount"], self["asset"]["precision"])
        return a

    def __mod__(self, other):
        a = self.copy()
        if isinstance(other, Amount):
            check_asset(other["asset"], self["asset"], self.blockchain)
            a["amount"] %= other["amount"]
        else:
            a["amount"] %= Decimal(other)
        if self.fixed_point_arithmetic:
            a["amount"] = quantize(a["amount"], self["asset"]["precision"])
        return a

    def __pow__(self, other):
        a = self.copy()
        if isinstance(other, Amount):
            check_asset(other["asset"], self["asset"], self.blockchain)
            a["amount"] **= other["amount"]
        else:
            a["amount"] **= Decimal(other)
        if self.fixed_point_arithmetic:
            a["amount"] = quantize(a["amount"], self["asset"]["precision"])
        return a

    def __iadd__(self, other):
        if isinstance(other, Amount):
            check_asset(other["asset"], self["asset"], self.blockchain)
            self["amount"] += other["amount"]
        else:
            self["amount"] += Decimal(other)
        if self.fixed_point_arithmetic:
            self["amount"] = quantize(self["amount"], self["asset"]["precision"])
        return self

    def __isub__(self, other):
        if isinstance(other, Amount):
            check_asset(other["asset"], self["asset"], self.blockchain)
            self["amount"] -= other["amount"]
        else:
            self["amount"] -= Decimal(other)
        if self.fixed_point_arithmetic:
            self["amount"] = quantize(self["amount"], self["asset"]["precision"])
        return self

    def __imul__(self, other):
        if isinstance(other, Amount):
            check_asset(other["asset"], self["asset"], self.blockchain)
            self["amount"] *= other["amount"]
        else:
            self["amount"] *= Decimal(other)

        self["amount"] = quantize(self["amount"], self["asset"]["precision"])
        return self

    def __idiv__(self, other):
        """
        In-place division: divide this Amount by another Amount or numeric value and return self.

        If `other` is an Amount, asserts asset compatibility and divides this object's internal amount by the other's amount. If `other` is numeric, divides by Decimal(other). When `fixed_point_arithmetic` is enabled, the result is quantized to this asset's precision.

        Returns:
            self (Amount): The mutated Amount instance.

        Raises:
            AssertionError: If `other` is an Amount with a different asset (via check_asset).
        """
        if isinstance(other, Amount):
            check_asset(other["asset"], self["asset"], self.blockchain)
            self["amount"] = self["amount"] / other["amount"]
        else:
            self["amount"] /= Decimal(other)
        if self.fixed_point_arithmetic:
            self["amount"] = quantize(self["amount"], self["asset"]["precision"])
        return self

    def __ifloordiv__(self, other):
        if isinstance(other, Amount):
            self["amount"] //= other["amount"]
        else:
            self["amount"] //= Decimal(other)
        self["amount"] = quantize(self["amount"], self["asset"]["precision"])
        return self

    def __imod__(self, other):
        if isinstance(other, Amount):
            check_asset(other["asset"], self["asset"], self.blockchain)
            self["amount"] %= other["amount"]
        else:
            self["amount"] %= Decimal(other)
        if self.fixed_point_arithmetic:
            self["amount"] = quantize(self["amount"], self["asset"]["precision"])
        return self

    def __ipow__(self, other):
        if isinstance(other, Amount):
            self["amount"] **= other
        else:
            self["amount"] **= Decimal(other)
        if self.fixed_point_arithmetic:
            self["amount"] = quantize(self["amount"], self["asset"]["precision"])
        return self

    def __lt__(self, other):
        quant_amount = quantize(self["amount"], self["asset"]["precision"])
        if isinstance(other, Amount):
            check_asset(other["asset"], self["asset"], self.blockchain)
            return quant_amount < quantize(other["amount"], self["asset"]["precision"])
        else:
            return quant_amount < quantize((other or 0), self["asset"]["precision"])

    def __le__(self, other):
        quant_amount = quantize(self["amount"], self["asset"]["precision"])
        if isinstance(other, Amount):
            check_asset(other["asset"], self["asset"], self.blockchain)
            return quant_amount <= quantize(other["amount"], self["asset"]["precision"])
        else:
            return quant_amount <= quantize((other or 0), self["asset"]["precision"])

    def __eq__(self, other):
        quant_amount = quantize(self["amount"], self["asset"]["precision"])
        if isinstance(other, Amount):
            check_asset(other["asset"], self["asset"], self.blockchain)
            return quant_amount == quantize(other["amount"], self["asset"]["precision"])
        else:
            return quant_amount == quantize((other or 0), self["asset"]["precision"])

    def __ne__(self, other):
        """
        Return True if this Amount is not equal to `other`.

        Compares values after quantizing both sides to this amount's asset precision. If `other` is an Amount, its asset must match this Amount's asset (an assertion is raised on mismatch) and the comparison uses both amounts quantized to the shared precision. If `other` is numeric or None, it is treated as a numeric value (None → 0) and compared after quantization.

        Returns:
                bool: True when the quantized values differ, False otherwise.
        """
        quant_amount = quantize(self["amount"], self["asset"]["precision"])
        if isinstance(other, Amount):
            check_asset(other["asset"], self["asset"], self.blockchain)
            return quantize(self["amount"], self["asset"]["precision"]) != quantize(
                other["amount"], self["asset"]["precision"]
            )
        else:
            return quant_amount != quantize((other or 0), self["asset"]["precision"])

    def __ge__(self, other):
        """
        Return True if this Amount is greater than or equal to `other`.

        Performs comparison after quantizing both values to this Amount's asset precision. If `other` is an Amount, its asset must match this Amount's asset (an AssertionError is raised on mismatch). If `other` is None, it is treated as zero. Returns a boolean.
        """
        quant_amount = quantize(self["amount"], self["asset"]["precision"])
        if isinstance(other, Amount):
            check_asset(other["asset"], self["asset"], self.blockchain)
            return quant_amount >= quantize(other["amount"], self["asset"]["precision"])
        else:
            return quant_amount >= quantize((other or 0), self["asset"]["precision"])

    def __gt__(self, other):
        quant_amount = quantize(self["amount"], self["asset"]["precision"])
        if isinstance(other, Amount):
            check_asset(other["asset"], self["asset"], self.blockchain)
            return quant_amount > quantize(other["amount"], self["asset"]["precision"])
        else:
            return quant_amount > quantize((other or 0), self["asset"]["precision"])

    __repr__ = __str__
    __truediv__ = __div__
    __itruediv__ = __idiv__
    __truemul__ = __mul__



from __future__ import absolute_import, division, print_function, unicode_literals

import re
from builtins import str


def decodeRPCErrorMsg(e):
    """Helper function to decode the raised Exception and give it a
    python Exception class
    """
    found = re.search(
        ("(10 assert_exception: Assert Exception\n|3030000 tx_missing_posting_auth).*: (.*)\n"),
        str(e),
        flags=re.M,
    )
    if found:
        return found.group(2).strip()
    else:
        return str(e)


class NectarApiException(Exception):
    """NectarApiException base Exception."""

    pass


class UnauthorizedError(NectarApiException):
    """UnauthorizedError Exception."""

    pass


class RPCConnection(NectarApiException):
    """RPCConnection Exception."""

    pass


class RPCError(NectarApiException):
    """RPCError Exception."""

    pass


class RPCErrorDoRetry(NectarApiException):
    """RPCErrorDoRetry Exception."""

    pass


class NumRetriesReached(NectarApiException):
    """NumRetriesReached Exception."""

    pass


class CallRetriesReached(NectarApiException):
    """CallRetriesReached Exception. Only for internal use"""

    pass


class MissingRequiredActiveAuthority(RPCError):
    pass


class UnknownKey(RPCError):
    pass


class NoMethodWithName(RPCError):
    pass


class NoApiWithName(RPCError):
    pass


class FollowApiNotEnabled(RPCError):
    pass


class ApiNotSupported(RPCError):
    pass


class UnhandledRPCError(RPCError):
    pass


class NoAccessApi(RPCError):
    pass


class FilteredItemNotFound(RPCError):
    pass


class InvalidEndpointUrl(NectarApiException):
    pass


class InvalidParameters(NectarApiException):
    pass


class SupportedByHivemind(NectarApiException):
    pass


class UnnecessarySignatureDetected(NectarApiException):
    pass


class WorkingNodeMissing(NectarApiException):
    pass


class TimeoutException(NectarApiException):
    pass


class VotedBeforeWaitTimeReached(NectarApiException):
    pass


class UnknownTransaction(NectarApiException):
    pass



# -*- coding: utf-8 -*-
# Inspired by https://raw.githubusercontent.com/xeroc/python-graphenelib/master/graphenestorage/base.py
import logging

from .exceptions import KeyAlreadyInStoreException
from .interfaces import (
    ConfigInterface,
    EncryptedKeyInterface,
    EncryptedTokenInterface,
    KeyInterface,
    TokenInterface,
)
from .masterpassword import MasterPassword
from .ram import InRamStore
from .sqlite import SQLiteStore

log = logging.getLogger(__name__)


# Configuration
class InRamConfigurationStore(InRamStore, ConfigInterface):
    """A simple example that stores configuration in RAM.

    Internally, this works by simply inheriting
    :class:`nectarstorage.ram.InRamStore`. The interface is defined in
    :class:`nectarstorage.interfaces.ConfigInterface`.
    """

    pass


class SqliteConfigurationStore(SQLiteStore, ConfigInterface):
    """This is the configuration storage that stores key/value
    pairs in the `config` table of the SQLite3 database.

    Internally, this works by simply inheriting
    :class:`nectarstorage.sqlite.SQLiteStore`. The interface is defined
    in :class:`nectarstorage.interfaces.ConfigInterface`.
    """

    #: The table name for the configuration
    __tablename__ = "config"
    #: The name of the 'key' column
    __key__ = "key"
    #: The name of the 'value' column
    __value__ = "value"


# Keys
class InRamPlainKeyStore(InRamStore, KeyInterface):
    """A simple in-RAM Store that stores keys unencrypted in RAM

    Internally, this works by simply inheriting
    :class:`nectarstorage.ram.InRamStore`. The interface is defined in
    :class:`nectarstorage.interfaces.KeyInterface`.
    """

    def getPublicKeys(self):
        return [k for k, v in self.items()]

    def getPrivateKeyForPublicKey(self, pub):
        return self.get(str(pub), None)

    def add(self, wif, pub):
        if str(pub) in self:
            raise KeyAlreadyInStoreException
        self[str(pub)] = str(wif)

    def delete(self, pub):
        InRamStore.delete(self, str(pub))


class SqlitePlainKeyStore(SQLiteStore, KeyInterface):
    """This is the key storage that stores the public key and the
    **unencrypted** private key in the `keys` table in the SQLite3
    database.

    Internally, this works by simply inheriting
    :class:`nectarstorage.ram.InRamStore`. The interface is defined in
    :class:`nectarstorage.interfaces.KeyInterface`.
    """

    #: The table name for the configuration
    __tablename__ = "keys"
    #: The name of the 'key' column
    __key__ = "pub"
    #: The name of the 'value' column
    __value__ = "wif"

    def getPublicKeys(self):
        return [k for k, v in self.items()]

    def getPrivateKeyForPublicKey(self, pub):
        return self[pub]

    def add(self, wif, pub):
        if str(pub) in self:
            raise KeyAlreadyInStoreException
        self[str(pub)] = str(wif)

    def delete(self, pub):
        SQLiteStore.delete(self, str(pub))

    def is_encrypted(self):
        """Returns False, as we are not encrypted here"""
        return False


class KeyEncryption(MasterPassword, EncryptedKeyInterface):
    """This is an interface class that provides the methods required for
    EncryptedKeyInterface and links them to the MasterPassword-provided
    functionatlity, accordingly.
    """

    def __init__(self, *args, **kwargs):
        EncryptedKeyInterface.__init__(self, *args, **kwargs)
        MasterPassword.__init__(self, *args, **kwargs)

    # Interface to deal with encrypted keys
    def getPublicKeys(self):
        return [k for k, v in self.items()]

    def getPrivateKeyForPublicKey(self, pub):
        wif = self.get(str(pub), None)
        if wif:
            return self.decrypt(wif)  # From Masterpassword

    def add(self, wif, pub):
        if str(pub) in self:
            raise KeyAlreadyInStoreException
        self[str(pub)] = self.encrypt(str(wif))  # From Masterpassword

    def is_encrypted(self):
        return True


class InRamEncryptedKeyStore(InRamStore, KeyEncryption):
    """An in-RAM Store that stores keys **encrypted** in RAM.

    Internally, this works by simply inheriting
    :class:`nectarstorage.ram.InRamStore`. The interface is defined in
    :class:`nectarstorage.interfaces.KeyInterface`.

    .. note:: This module also inherits
        :class:`nectarstorage.masterpassword.MasterPassword` which offers
        additional methods and deals with encrypting the keys.
    """

    def __init__(self, *args, **kwargs):
        InRamStore.__init__(self, *args, **kwargs)
        KeyEncryption.__init__(self, *args, **kwargs)


class SqliteEncryptedKeyStore(SQLiteStore, KeyEncryption):
    """This is the key storage that stores the public key and the
    **encrypted** private key in the `keys` table in the SQLite3 database.

    Internally, this works by simply inheriting
    :class:`nectarstorage.ram.InRamStore`. The interface is defined in
    :class:`nectarstorage.interfaces.KeyInterface`.

    .. note:: This module also inherits
        :class:`nectarstorage.masterpassword.MasterPassword` which offers
        additional methods and deals with encrypting the keys.
    """

    __tablename__ = "keys"
    __key__ = "pub"
    __value__ = "wif"

    def __init__(self, *args, **kwargs):
        SQLiteStore.__init__(self, *args, **kwargs)
        KeyEncryption.__init__(self, *args, **kwargs)


# Token
class InRamPlainTokenStore(InRamStore, TokenInterface):
    """A simple in-RAM Store that stores token unencrypted in RAM

    Internally, this works by simply inheriting
    :class:`nectarstorage.ram.InRamStore`. The interface is defined in
    :class:`nectarstorage.interfaces.TokenInterface`.
    """

    def getPublicNames(self):
        return [k for k, v in self.items()]

    def getPrivateKeyForPublicKey(self, pub):
        return self.get(str(pub), None)

    def add(self, token, name):
        if str(name) in self:
            raise KeyAlreadyInStoreException
        self[str(name)] = str(token)

    def delete(self, name):
        InRamStore.delete(self, str(name))


class SqlitePlainTokenStore(SQLiteStore, TokenInterface):
    """This is the token storage that stores the public key and the
    **unencrypted** private key in the `tokens` table in the SQLite3
    database.

    Internally, this works by simply inheriting
    :class:`nectarstorage.ram.InRamStore`. The interface is defined in
    :class:`nectarstorage.interfaces.TokenInterface`.
    """

    #: The table name for the configuration
    __tablename__ = "token"
    #: The name of the 'key' column
    __key__ = "name"
    #: The name of the 'value' column
    __value__ = "token"

    def getPublicNames(self):
        return [k for k, v in self.items()]

    def getPrivateKeyForPublicKey(self, name):
        return self[name]

    def add(self, token, name):
        if str(name) in self:
            raise KeyAlreadyInStoreException
        self[str(name)] = str(token)

    def updateToken(self, name, token):
        self[str(name)] = str(token)  # From Masterpassword

    def delete(self, name):
        SQLiteStore.delete(self, str(name))

    def is_encrypted(self):
        """Returns False, as we are not encrypted here"""
        return False


class TokenEncryption(MasterPassword, EncryptedTokenInterface):
    """This is an interface class that provides the methods required for
    EncryptedTokenInterface and links them to the MasterPassword-provided
    functionatlity, accordingly.
    """

    def __init__(self, *args, **kwargs):
        EncryptedTokenInterface.__init__(self, *args, **kwargs)
        MasterPassword.__init__(self, *args, **kwargs)

    # Interface to deal with encrypted keys
    def getPublicNames(self):
        return [k for k, v in self.items()]

    def getPrivateKeyForPublicKey(self, name):
        token = self.get(str(name), None)
        if token:
            return self.decrypt_text(token)  # From Masterpassword

    def add(self, token, name):
        if str(name) in self:
            raise KeyAlreadyInStoreException
        self[str(name)] = self.encrypt_text(str(token))  # From Masterpassword

    def updateToken(self, name, token):
        self[str(name)] = self.encrypt_text(str(token))  # From Masterpassword

    def is_encrypted(self):
        return True


class InRamEncryptedTokenStore(InRamStore, TokenEncryption):
    """An in-RAM Store that stores token **encrypted** in RAM.

    Internally, this works by simply inheriting
    :class:`nectarstorage.ram.InRamStore`. The interface is defined in
    :class:`nectarstorage.interfaces.TokenInterface`.

    .. note:: This module also inherits
        :class:`nectarstorage.masterpassword.MasterPassword` which offers
        additional methods and deals with encrypting the keys.
    """

    def __init__(self, *args, **kwargs):
        InRamStore.__init__(self, *args, **kwargs)
        TokenEncryption.__init__(self, *args, **kwargs)


class SqliteEncryptedTokenStore(SQLiteStore, TokenEncryption):
    """This is the key storage that stores the account name and the
    **encrypted** token in the `token` table in the SQLite3 database.

    Internally, this works by simply inheriting
    :class:`nectarstorage.ram.InRamStore`. The interface is defined in
    :class:`nectarstorage.interfaces.TokenInterface`.

    .. note:: This module also inherits
        :class:`nectarstorage.masterpassword.MasterPassword` which offers
        additional methods and deals with encrypting the token.
    """

    __tablename__ = "token"
    __key__ = "name"
    __value__ = "token"

    def __init__(self, *args, **kwargs):
        SQLiteStore.__init__(self, *args, **kwargs)
        TokenEncryption.__init__(self, *args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from datetime import timedelta
from timeit import default_timer as timer

from prettytable import PrettyTable

from nectar.account import Account
from nectar.block import Block
from nectar.blockchain import Blockchain
from nectar.comment import Comment
from nectar.hive import Hive
from nectar.nodelist import NodeList
from nectar.utils import (
    construct_authorperm,
    parse_time,
    resolve_authorpermvoter,
)
from nectar.vote import Vote
from nectarapi.exceptions import NumRetriesReached

FUTURES_MODULE = None
if not FUTURES_MODULE:
    try:
        from concurrent.futures import ThreadPoolExecutor, as_completed

        FUTURES_MODULE = "futures"
    except ImportError:
        FUTURES_MODULE = None
log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)
quit_thread = False


def benchmark_node(node, how_many_minutes=10, how_many_seconds=30):
    """
    Benchmark a single Hive node by measuring block processing, account history retrieval, and basic RPC access latencies.

    Performs three timed phases against the provided node:
    1. Iterates historical blocks starting from a fixed block ID until a time limit (how_many_minutes) or per-call time limit (how_many_seconds) is reached, counting blocks and transactions.
    2. Iterates the account history of the fixed account "gtg" in reverse (batch_size=100) until the per-call time limit is reached, counting history entries.
    3. Measures the average latency of four representative operations (Vote, Comment, Account construction, and fetching followers) to produce an access-time metric.

    Parameters:
        node (str): URL or identifier of the Hive node to benchmark.
        how_many_minutes (int): Maximum span (in minutes) of blockchain history to walk from the starting block (default: 10).
        how_many_seconds (int): Per-phase timeout in seconds to abort a long-running loop/scan (default: 30).

    Returns:
        dict: A summary with the following keys:
            - successful (bool): False if a retriable error occurred during history or access phases.
            - node (str): The node value passed in.
            - error (str|None): Error message when a phase failed, otherwise None.
            - total_duration (float): Total elapsed seconds for the entire benchmark run.
            - block_count (int): Number of blocks processed (or -1 on error).
            - history_count (int): Number of account history entries processed (or -1 on error).
            - access_time (float): Average latency (seconds) of the measured RPC operations (or -1 on error).
            - follow_time (float): Measured time (seconds) for fetching followers in the access phase.
            - version (str): Blockchain version string reported by the node.
    """
    block_count = 0
    history_count = 0
    access_time = 0
    follow_time = 0
    blockchain_version = "0.0.0"
    successful = True
    error_msg = None
    start_total = timer()
    max_batch_size = None
    threading = False
    thread_num = 16

    authorpermvoter = "@gtg/hive-pressure-4-need-for-speed|gandalf"
    [author, permlink, voter] = resolve_authorpermvoter(authorpermvoter)
    authorperm = construct_authorperm(author, permlink)
    last_block_id = 19273700
    try:
        hv = Hive(node=node, num_retries=3, num_retries_call=3, timeout=30)
        blockchain = Blockchain(blockchain_instance=hv)
        blockchain_version = hv.get_blockchain_version()

        last_block = Block(last_block_id, blockchain_instance=hv)

        stopTime = last_block.time() + timedelta(seconds=how_many_minutes * 60)
        total_transaction = 0

        start = timer()
        for entry in blockchain.blocks(
            start=last_block_id,
            max_batch_size=max_batch_size,
            threading=threading,
            thread_num=thread_num,
        ):
            block_no = entry.identifier
            block_count += 1
            if "block" in entry:
                trxs = entry["block"]["transactions"]
            else:
                trxs = entry["transactions"]

            for tx in trxs:
                for op in tx["operations"]:
                    total_transaction += 1
            if "block" in entry:
                block_time = parse_time(entry["block"]["timestamp"])
            else:
                block_time = parse_time(entry["timestamp"])

            if block_time > stopTime:
                last_block_id = block_no
                break
            if timer() - start > how_many_seconds or quit_thread:
                break
    except NumRetriesReached:
        error_msg = "NumRetriesReached"
        block_count = -1
    except KeyboardInterrupt:
        error_msg = "KeyboardInterrupt"
        # quit = True
    except Exception as e:
        error_msg = str(e)
        block_count = -1

    try:
        hv = Hive(node=node, num_retries=3, num_retries_call=3, timeout=30)
        account = Account("gtg", blockchain_instance=hv)
        blockchain_version = hv.get_blockchain_version()

        start = timer()
        for acc_op in account.history_reverse(batch_size=100):
            history_count += 1
            if timer() - start > how_many_seconds or quit_thread:
                break
    except NumRetriesReached:
        error_msg = "NumRetriesReached"
        history_count = -1
        successful = False
    except KeyboardInterrupt:
        error_msg = "KeyboardInterrupt"
        history_count = -1
        successful = False
        # quit = True
    except Exception as e:
        error_msg = str(e)
        history_count = -1
        successful = False

    try:
        hv = Hive(node=node, num_retries=3, num_retries_call=3, timeout=30)
        account = Account("gtg", blockchain_instance=hv)
        blockchain_version = hv.get_blockchain_version()

        start = timer()
        Vote(authorpermvoter, blockchain_instance=hv)
        stop = timer()
        vote_time = stop - start
        start = timer()
        Comment(authorperm, blockchain_instance=hv)
        stop = timer()
        comment_time = stop - start
        start = timer()
        Account(author, blockchain_instance=hv)
        stop = timer()
        account_time = stop - start
        start = timer()
        account.get_followers()
        stop = timer()
        follow_time = stop - start
        access_time = (vote_time + comment_time + account_time + follow_time) / 4.0
    except NumRetriesReached:
        error_msg = "NumRetriesReached"
        access_time = -1
    except KeyboardInterrupt:
        error_msg = "KeyboardInterrupt"
        # quit = True
    except Exception as e:
        error_msg = str(e)
        access_time = -1
    return {
        "successful": successful,
        "node": node,
        "error": error_msg,
        "total_duration": timer() - start_total,
        "block_count": block_count,
        "history_count": history_count,
        "access_time": access_time,
        "follow_time": follow_time,
        "version": blockchain_version,
    }


if __name__ == "__main__":
    how_many_seconds = 30
    how_many_minutes = 10
    threading = True
    set_default_nodes = False
    quit_thread = False
    benchmark_time = timer()

    nodelist = NodeList()
    nodes = nodelist.get_nodes(normal=True, appbase=True, dev=True)
    t = PrettyTable(["node", "N blocks", "N acc hist", "dur. call in s"])
    t.align = "l"
    t2 = PrettyTable(["node", "version"])
    t2.align = "l"
    working_nodes = []
    results = []
    if threading and FUTURES_MODULE:
        pool = ThreadPoolExecutor(max_workers=len(nodes) + 1)
        futures = []
        for node in nodes:
            futures.append(pool.submit(benchmark_node, node, how_many_minutes, how_many_seconds))
        try:
            results = [r.result() for r in as_completed(futures)]
        except KeyboardInterrupt:
            quit_thread = True
            print("benchmark aborted.")
    else:
        for node in nodes:
            print("Current node:", node)
            result = benchmark_node(node, how_many_minutes, how_many_seconds)
            results.append(result)
    for result in results:
        t2.add_row([result["node"], result["version"]])
    print(t2)
    print("\n")

    sortedList = sorted(results, key=lambda self: self["history_count"], reverse=True)
    for result in sortedList:
        if result["successful"]:
            t.add_row(
                [
                    result["node"],
                    result["block_count"],
                    result["history_count"],
                    ("%.2f" % (result["access_time"])),
                ]
            )
            working_nodes.append(result["node"])
    print(t)
    print("\n")
    print("Total benchmark time: %.2f s\n" % (timer() - benchmark_time))
    if set_default_nodes:
        hv = Hive(offline=True)
        hv.set_default_nodes(working_nodes)
    else:
        print("hive-nectar set nodes " + str(working_nodes))



# -*- coding: utf-8 -*-
import nectar


class SharedInstance(object):
    """Singleton for the shared Blockchain Instance (Hive-only)."""

    instance = None
    config = {}


def shared_blockchain_instance():
    """Initialize and return the shared Hive instance.

    Hive-only: this always returns a `nectar.Hive` instance, regardless of any
    legacy configuration that may have referenced other chains.
    """
    if not SharedInstance.instance:
        clear_cache()
        SharedInstance.instance = nectar.Hive(**SharedInstance.config)
    return SharedInstance.instance


def set_shared_blockchain_instance(blockchain_instance):
    """
    Override the shared Hive instance used by the module and clear related caches.

    This sets SharedInstance.instance to the provided blockchain instance and calls clear_cache()
    to invalidate any cached blockchain objects so consumers observe the new instance immediately.
    """
    clear_cache()
    SharedInstance.instance = blockchain_instance


def shared_hive_instance():
    """Initialize (if needed) and return the shared Hive instance."""
    return shared_blockchain_instance()


def set_shared_hive_instance(hive_instance):
    """
    Override the global shared Hive instance used by the module.

    Replaces the current SharedInstance.instance with the provided hive_instance and clears related caches so subsequent calls return the new instance.

    Parameters:
        hive_instance: The nectar.Hive instance to set as the shared global instance.
    """
    set_shared_blockchain_instance(hive_instance)


def clear_cache():
    """
    Clear cached blockchain object state.

    Performs a lazy import of BlockchainObject and calls its clear_cache() method to purge any in-memory caches of blockchain objects (used when the shared Hive instance or configuration changes).
    """
    from .blockchainobject import BlockchainObject

    BlockchainObject.clear_cache()


def set_shared_config(config):
    """
    Set configuration for the shared Hive instance without creating the instance.

    Updates the global SharedInstance.config with the provided mapping. If a shared instance already exists, clears internal caches and resets the shared instance to None so the new configuration will take effect on next access.

    Parameters:
        config (dict): Configuration options to merge into the shared instance configuration.

    Raises:
        AssertionError: If `config` is not a dict.
    """
    if not isinstance(config, dict):
        raise AssertionError()
    SharedInstance.config.update(config)
    # if one is already set, delete
    if SharedInstance.instance:
        clear_cache()
        SharedInstance.instance = None
